import construct as cs
from . import GalleryItem


constr = cs.Struct(
    "len" / cs.Int8ul,
    "bytes_fix" / cs.Bytes(5),
    "bytes_lambda" / cs.Bytes(lambda ctx: 2),
    "bytes_this" / cs.Bytes(cs.this.len),
    "greedybytes" / cs.GreedyBytes,
)


gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Zeros": bytes([3]) + bytes(15),
        "1": b"\x03123456789ABCDEF",
    },
)


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


constr = cs.Struct(
    "brand" / cs.Enum(cs.Int8ul, Porsche=0, Audi=4, VW=7),
    "wheels" / cs.Int8ul,
    "color" / cs.Enum(cs.Int8ul, Red=1, Green=10, Blue=11, Black=12),
    "long_list"
    / cs.Enum(
        cs.Int8ul,
        Entry0=0,
        Entry1=1,
        Entry2=2,
        Entry3=3,
        Entry4=4,
        Entry5=5,
        Entry6=6,
        Entry7=7,
        Entry8=8,
        Entry9=9,
        Entry10=10,
        Entry11=11,
        Entry12=12,
        Entry13=13,
        Entry14=14,
        Entry15=15,
        Entry16=16,
        Entry17=17,
        Entry18=18,
        Entry19=19,
        Entry20=20,
        Entry21=21,
        Entry22=22,
        Entry23=23,
        Entry24=24,
        Entry25=25,
        Entry26=26,
        Entry27=27,
        Entry28=28,
        Entry29=29,
        Entry30=30,
        Entry31=31,
        Entry32=32,
        Entry33=33,
        Entry34=34,
        Entry35=35,
        Entry36=36,
        Entry37=37,
        Entry38=38,
        Entry39=39,
        Entry40=40,
        Entry41=41,
        Entry42=42,
        Entry43=43,
        Entry44=44,
        Entry45=45,
        Entry46=46,
        Entry47=47,
        Entry48=48,
        Entry49=49,
        Entry50=50,
    ),
)


gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "3": bytes([7, 2, 1, 7]),
        "2": bytes([4, 4, 13, 6]),
        "1": bytes([4, 4, 12, 5]),
        "Zeros": bytes(constr.sizeof()),
    },
)


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


@dataclasses.dataclass
class SwitchTest(cst.DataclassMixin):
    @dataclasses.dataclass
    class Case1(cst.DataclassMixin):
        case1_1: int = cst.csfield(cs.Int16sb)
        case1_2: int = cst.csfield(cs.Int16sb)

    @dataclasses.dataclass
    class Case2(cst.DataclassMixin):
        case2_1: int = cst.csfield(cs.Int8sb)
        case2_2: int = cst.csfield(cs.Int8sb)
        case2_3: int = cst.csfield(cs.Int8sb)
        case2_4: int = cst.csfield(cs.Int8sb)

    @dataclasses.dataclass
    class CaseDefault(cst.DataclassMixin):
        case_default_1: int = cst.csfield(cs.Int32sb)

    choice: int = cst.csfield(cs.Int8ub)
    switch: t.Union[Case1, Case2, CaseDefault] = cst.csfield(
        cs.Switch(
            cs.this.choice,
            cases={
                1: cst.DataclassStruct(Case1),
                2: cst.DataclassStruct(Case2),
            },
            default=cst.DataclassStruct(CaseDefault),
        )
    )

    switch_without_default: t.Union[Case1, Case2, None] = cst.csfield(
        cs.Switch(
            cs.this.choice,
            cases={
                1: cst.DataclassStruct(Case1),
                2: cst.DataclassStruct(Case2),
            },
            default=cs.Pass,
        )
    )


constr = cst.DataclassStruct(SwitchTest)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Default": bytes([0, 0, 0, 0, 0]),
        "Case 1": bytes([1, 1, 2, 1, 2, 5, 6, 7, 8]),
        "Case 2": bytes([2, 1, 2, 1, 2, 5, 6, 7, 8]),
    },
)


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


constr = cs.Struct(
    "permissions" / cs.FlagsEnum(cs.Int8ul, R=4, W=2, X=1),
    "days"
    / cs.FlagsEnum(
        cs.Int8ul,
        Monday=(1 << 0),
        Tuesday=(1 << 1),
        Wednesday=(1 << 2),
        Thursday=(1 << 3),
        Friday=(1 << 4),
        Saturday=(1 << 5),
        Sunday=(1 << 6),
    ),
    "long_list"
    / cs.FlagsEnum(
        cs.Int64ul,
        Entry0=(1 << 0),
        Entry1=(1 << 1),
        Entry2=(1 << 2),
        Entry3=(1 << 3),
        Entry4=(1 << 4),
        Entry5=(1 << 5),
        Entry6=(1 << 6),
        Entry7=(1 << 7),
        Entry8=(1 << 8),
        Entry9=(1 << 9),
        Entry10=(1 << 10),
        Entry11=(1 << 11),
        Entry12=(1 << 12),
        Entry13=(1 << 13),
        Entry14=(1 << 14),
        Entry15=(1 << 15),
        Entry16=(1 << 16),
        Entry17=(1 << 17),
        Entry18=(1 << 18),
        Entry19=(1 << 19),
        Entry20=(1 << 20),
        Entry21=(1 << 21),
        Entry22=(1 << 22),
        Entry23=(1 << 23),
        Entry24=(1 << 24),
        Entry25=(1 << 25),
        Entry26=(1 << 26),
        Entry27=(1 << 27),
        Entry28=(1 << 28),
        Entry29=(1 << 29),
        Entry30=(1 << 30),
        Entry31=(1 << 31),
        Entry32=(1 << 32),
        Entry33=(1 << 33),
        Entry34=(1 << 34),
        Entry35=(1 << 35),
        Entry36=(1 << 36),
        Entry37=(1 << 37),
        Entry38=(1 << 38),
        Entry39=(1 << 39),
        Entry40=(1 << 40),
        Entry41=(1 << 41),
        Entry42=(1 << 42),
        Entry43=(1 << 43),
        Entry44=(1 << 44),
        Entry45=(1 << 45),
        Entry46=(1 << 46),
        Entry47=(1 << 47),
        Entry48=(1 << 48),
        Entry49=(1 << 49),
        Entry50=(1 << 50),
    ),
)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "2": constr.build(
            {
                "permissions": constr.permissions.R | constr.permissions.W,
                "days": constr.days.Monday | constr.days.Sunday,
                "long_list": constr.long_list.Entry49 | constr.long_list.Entry50,
            }
        ),
        "1": constr.build(
            {
                "permissions": constr.permissions.R,
                "days": constr.days.Monday,
                "long_list": constr.long_list.Entry0,
            }
        ),
        "Zeros": constr.build(
            {
                "permissions": 0,
                "days": 0,
                "long_list": 0,
            }
        ),
    },
)



import construct as cs
from . import GalleryItem


constr = cs.Struct(
    "padded" / cs.Padded(5, cs.Bytes(3)),
    "padding" / cs.Padding(5),
)


gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "1": bytes([0, 1, 2, 3, 4, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5]),
        "Zeros": bytes(10),
    },
)


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


class CarBrand(cst.EnumBase):
    Porsche = 0
    Audi = 4
    VW = 7


class CarColor(cst.EnumBase):
    Red = 1
    Green = 10
    Blue = 11
    Black = 12


class LongList(cst.EnumBase):
    Entry0 = 0
    Entry1 = 1
    Entry2 = 2
    Entry3 = 3
    Entry4 = 4
    Entry5 = 5
    Entry6 = 6
    Entry7 = 7
    Entry8 = 8
    Entry9 = 9
    Entry10 = 10
    Entry11 = 11
    Entry12 = 12
    Entry13 = 13
    Entry14 = 14
    Entry15 = 15
    Entry16 = 16
    Entry17 = 17
    Entry18 = 18
    Entry19 = 19
    Entry20 = 20
    Entry21 = 21
    Entry22 = 22
    Entry23 = 23
    Entry24 = 24
    Entry25 = 25
    Entry26 = 26
    Entry27 = 27
    Entry28 = 28
    Entry29 = 29
    Entry30 = 30
    Entry31 = 31
    Entry32 = 32
    Entry33 = 33
    Entry34 = 34
    Entry35 = 35
    Entry36 = 36
    Entry37 = 37
    Entry38 = 38
    Entry39 = 39
    Entry40 = 40
    Entry41 = 41
    Entry42 = 42
    Entry43 = 43
    Entry44 = 44
    Entry45 = 45
    Entry46 = 46
    Entry47 = 47
    Entry48 = 48
    Entry49 = 49
    Entry50 = 50


@dataclasses.dataclass
class Car(cst.DataclassMixin):
    brand: CarBrand = cst.csfield(cst.TEnum(cs.Int8ul, CarBrand))
    wheels: int = cst.csfield(cs.Int8ul)
    color: CarColor = cst.csfield(cst.TEnum(cs.Int8ul, CarColor))
    long_list: LongList = cst.csfield(cst.TEnum(cs.Int8ul, LongList))


constr = cst.DataclassStruct(Car)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "3": bytes([7, 2, 1, 7]),
        "2": bytes([4, 4, 13, 6]),
        "1": bytes([4, 4, 12, 5]),
        "Zeros": bytes(constr.sizeof()),
    },
)


# -*- coding: utf-8 -*-
from typing import Any, Optional

import construct as cs
import wx

from construct_editor.widgets.construct_editor import ConstructEditor
from construct_editor.widgets.hex_editor import (
    HexEditor,
    HexEditorFormat,
    HexEditorBinaryData,
)
from construct_editor.helper.wrapper import (
    EntryConstruct,
)


class ConstructHexEditor(wx.Panel):
    def __init__(
        self,
        parent,
        construct: cs.Construct,
        contextkw: dict = {},
        binary: Optional[bytes] = None,
    ):
        super().__init__(parent)

        self._contextkw = contextkw

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        # create HexEditor
        self.hex_editor = HexEditor(
            self,
            binary,
            HexEditorFormat(width=16),
        )
        hsizer.Add(self.hex_editor, 0, wx.EXPAND | wx.ALL, 5)
        self.hex_editor.on_binary_changed.append(
            lambda _: self._convert_binary_to_struct()
        )

        # add button as line to toggle the visibility of the HexEditor
        self.toggle_hex_visibility_btn = wx.Button(
            self, wx.ID_ANY, "«", size=wx.Size(12, -1)
        )  # »
        hsizer.Add(self.toggle_hex_visibility_btn, 0, wx.EXPAND | wx.ALL, 0)
        self.toggle_hex_visibility_btn.Bind(
            wx.EVT_BUTTON, lambda evt: self.toggle_hex_visibility()
        )

        # create ConstructEditor
        self.construct_editor = ConstructEditor(
            self,
            construct,
        )
        hsizer.Add(self.construct_editor, 1, wx.ALL | wx.EXPAND, 5)
        self.construct_editor.on_root_obj_changed.append(
            lambda _: self._convert_struct_to_binary()
        )
        self.construct_editor.on_entry_selected.append(self._on_entry_selected)

        self._converting = False
        self._hex_editor_visible = True

        # show data in construct editor
        self.refresh()

        self.SetSizer(hsizer)
        self.Layout()

    def refresh(self):
        """ Refresh the content of the construct view """
        self.Freeze()
        self.hex_editor.refresh()
        self._convert_binary_to_struct()
        self.Thaw()

    def toggle_hex_visibility(self):
        """ Toggle the visibility of the HexEditor """
        if self._hex_editor_visible:
            self.hex_editor.HideWithEffect(wx.SHOW_EFFECT_ROLL_TO_LEFT)
            self.toggle_hex_visibility_btn.SetLabelText("»")
            self._hex_editor_visible = False
        else:
            self.hex_editor.ShowWithEffect(wx.SHOW_EFFECT_ROLL_TO_RIGHT)
            self.toggle_hex_visibility_btn.SetLabelText("«")
            self._hex_editor_visible = True
        self.Freeze()
        self.Layout()
        self.Refresh()
        self.Thaw()

    # Property: construct #####################################################
    @property
    def construct(self) -> cs.Construct:
        """ construct used for parsing """
        return self.construct_editor.construct

    @construct.setter
    def construct(self, val: cs.Construct):
        self.construct_editor.construct = val

    # Property: contextkw #####################################################
    @property
    def contextkw(self) -> dict:
        """ contextkw used for parsing the construct """
        return self._contextkw

    @contextkw.setter
    def contextkw(self, val: dict):
        self._contextkw = val

    # Property: root_obj ######################################################
    @property
    def root_obj(self) -> Any:
        return self.construct_editor.root_obj

    # Property: binary ########################################################
    @property
    def binary(self) -> bytes:
        return self.hex_editor.binary

    @binary.setter
    def binary(self, val: bytes):
        self.hex_editor.binary = val

    # Internals ###############################################################
    def _convert_binary_to_struct(self):
        """ Convert binary to construct object """
        if self._converting:
            return
        try:
            self._converting = True
            self.construct_editor.parse(self.hex_editor.binary, **self.contextkw)
        finally:
            self._converting = False

    def _convert_struct_to_binary(self):
        """ Convert construct object to binary """
        try:
            self._converting = True
            binary = self.construct_editor.build(**self.contextkw)
            self.hex_editor.binary = binary
        except Exception:
            pass  # ignore errors, because they are already shown in the gui
        finally:
            self._converting = False

    def _on_entry_selected(self, entry: EntryConstruct):
        metadata = entry.obj_metadata
        if metadata is not None:
            self.hex_editor.colorise(metadata.offset_start, metadata.offset_end, refresh=False)
            self.hex_editor.scroll_to_idx(metadata.offset_end - 1, refresh=False)
            self.hex_editor.scroll_to_idx(metadata.offset_start, refresh=False)
            self.hex_editor.refresh()
        else:
            self.hex_editor.colorise(0, 0)



from __future__ import annotations

import dataclasses
import sys
import typing as t

from pathlib import Path
import construct as cs
import wx
from wx.lib.embeddedimage import PyEmbeddedImage

import construct_editor.gallery.example_pe32coff
import construct_editor.gallery.example_ipstack
import construct_editor.gallery.test_bytes_greedybytes
import construct_editor.gallery.test_array
import construct_editor.gallery.test_greedyrange
import construct_editor.gallery.test_renamed
import construct_editor.gallery.test_ifthenelse
import construct_editor.gallery.test_switch
import construct_editor.gallery.test_dataclass_struct
import construct_editor.gallery.test_dataclass_bit_struct
import construct_editor.gallery.test_enum
import construct_editor.gallery.test_flagsenum
import construct_editor.gallery.test_tenum
import construct_editor.gallery.test_tflagsenum
import construct_editor.gallery.test_computed
import construct_editor.gallery.test_timestamp
import construct_editor.gallery.test_padded
import construct_editor.gallery.test_aligned
import construct_editor.gallery.test_pointer_peek_seek_tell
import construct_editor.gallery.test_pass
from construct_editor.widgets.construct_hex_editor import ConstructHexEditor


class ConstructGalleryFrame(wx.Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.SetTitle("Construct Gallery")
        self.SetSize(1600, 1000)
        self.SetIcon(icon.GetIcon())
        self.Center()

        self.main_panel = ConstructGallery(self)

        self.status_bar: wx.StatusBar = self.CreateStatusBar()


class ConstructGallery(wx.Panel):
    def __init__(self, parent: ConstructGalleryFrame):
        super().__init__(parent)

        # Define all galleries ############################################
        self.construct_gallery = {
            "################ EXAMPLES ################": None,
            "Example: pe32coff": construct_editor.gallery.example_pe32coff.gallery_item,
            "Example: ipstack": construct_editor.gallery.example_ipstack.gallery_item,
            "################ TESTS ####################": None,
            # "## bytes and bits ################": None,
            "Test: Bytes/GreedyBytes": construct_editor.gallery.test_bytes_greedybytes.gallery_item,
            # "## integers and floats ###########": None,
            # "Test: FormatField (TODO)": None,
            # "Test: BytesInteger (TODO)": None,
            # "Test: BitsInteger (TODO)": None,
            # "## strings #######################": None,
            # "Test: StringEncoded (TODO)": None,
            "## mappings ######################": None,
            # "Test: Flag (TODO)": None,
            "Test: Enum": construct_editor.gallery.test_enum.gallery_item,
            "Test: FlagsEnum": construct_editor.gallery.test_flagsenum.gallery_item,
            "Test: TEnum": construct_editor.gallery.test_tenum.gallery_item,
            "Test: TFlagsEnum": construct_editor.gallery.test_tflagsenum.gallery_item,
            # "Test: Mapping (TODO)": None,
            "## structures and sequences ######": None,
            # "Test: Struct (TODO)": None,
            # "Test: Sequence (TODO)": None,
            "Test: DataclassStruct": construct_editor.gallery.test_dataclass_struct.gallery_item,
            "Test: DataclassBitStruct": construct_editor.gallery.test_dataclass_bit_struct.gallery_item,
            "## arrays ranges and repeaters ######": None,
            "Test: Array": construct_editor.gallery.test_array.gallery_item,
            "Test: GreedyRange": construct_editor.gallery.test_greedyrange.gallery_item,
            # "Test: RepeatUntil (TODO)": None,
            "## specials ##########################": None,
            "Test: Renamed": construct_editor.gallery.test_renamed.gallery_item,
            "## miscellaneous ##########################": None,
            # "Test: Const (TODO)": None,
            "Test: Computed": construct_editor.gallery.test_computed.gallery_item,
            # "Test: Index (TODO)": None,
            # "Test: Rebuild (TODO)": None,
            # "Test: Default (TODO)": None,
            # "Test: Check (TODO)": None,
            # "Test: Error (TODO)": None,
            # "Test: FocusedSeq (TODO)": None,
            # "Test: Pickled (TODO)": None,
            # "Test: Numpy (TODO)": None,
            # "Test: NamedTuple (TODO)": None,
            "Test: TimestampAdapter": construct_editor.gallery.test_timestamp.gallery_item,
            # "Test: Hex (TODO)": None,
            # "Test: HexDump (TODO)": None,
            "## conditional ##########################": None,
            # "Test: Union (TODO)": None,
            # "Test: Select (TODO)": None,
            "Test: IfThenElse": construct_editor.gallery.test_ifthenelse.gallery_item,
            "Test: Switch": construct_editor.gallery.test_switch.gallery_item,
            # "Test: StopIf (TODO)": None,
            "## alignment and padding ##########################": None,
            "Test: Padded": construct_editor.gallery.test_padded.gallery_item,
            "Test: Aligned": construct_editor.gallery.test_aligned.gallery_item,
            "## stream manipulation ##########################": None,
            "Test: Pointer/Peek/Seek/Tell": construct_editor.gallery.test_pointer_peek_seek_tell.gallery_item,
            "Test: Pass": construct_editor.gallery.test_pass.gallery_item,
            # "Test: Terminated (TODO)": None,
            # "## tunneling and byte/bit swapping ##########################": None,
            # "Test: RawCopy (TODO)": None,
            # "Test: Prefixed (TODO)": None,
            # "Test: FixedSized (TODO)": None,
            # "Test: NullTerminated (TODO)": None,
            # "Test: NullStripped (TODO)": None,
            # "Test: RestreamData (TODO)": None,
            # "Test: Transformed (TODO)": None,
            # "Test: Restreamed (TODO)": None,
            # "Test: ProcessXor (TODO)": None,
            # "Test: ProcessRotateLeft (TODO)": None,
            # "Test: Checksum (TODO)": None,
            # "Test: Compressed (TODO)": None,
            # "Test: CompressedLZ4 (TODO)": None,
            # "Test: Rebuffered (TODO)": None,
        }
        self.gallery_selection = 1
        default_gallery = list(self.construct_gallery.keys())[self.gallery_selection]
        default_gallery_item = self.construct_gallery[default_gallery]

        # Define GUI elements #############################################
        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        vsizer = wx.BoxSizer(wx.VERTICAL)

        # gallery selctor
        self.gallery_selector_lbx = wx.ListBox(
            self,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.DefaultSize,
            list(self.construct_gallery.keys()),
            0,
            name="gallery_selector",
        )
        self.gallery_selector_lbx.SetStringSelection(default_gallery)
        vsizer.Add(self.gallery_selector_lbx, 1, wx.ALL, 1)

        # reload btn
        self.reload_btn = wx.Button(
            self, wx.ID_ANY, "Reload", wx.DefaultPosition, wx.DefaultSize, 0
        )
        vsizer.Add(self.reload_btn, 0, wx.ALL | wx.EXPAND, 1)

        # line
        vsizer.Add(wx.StaticLine(self), 0, wx.TOP | wx.BOTTOM | wx.EXPAND, 5)

        # clear binary
        self.clear_binary_btn = wx.Button(
            self, wx.ID_ANY, "Clear Binary", wx.DefaultPosition, wx.DefaultSize, 0
        )
        vsizer.Add(self.clear_binary_btn, 0, wx.ALL | wx.EXPAND, 1)

        # example selctor
        self.example_selector_lbx = wx.ListBox(
            self,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.Size(-1, 100),
            list(default_gallery_item.example_binarys.keys()),
            0,
            name="gallery_selector",
        )
        if len(default_gallery_item.example_binarys) > 0:
            self.example_selector_lbx.SetStringSelection(
                list(default_gallery_item.example_binarys.keys())[0]
            )
        vsizer.Add(self.example_selector_lbx, 0, wx.ALL | wx.EXPAND, 1)

        # load binary from file
        self.load_binary_file_btn = wx.Button(
            self,
            wx.ID_ANY,
            "Load Binary from File",
            wx.DefaultPosition,
            wx.DefaultSize,
            0,
        )
        vsizer.Add(self.load_binary_file_btn, 0, wx.ALL | wx.EXPAND, 1)

        self.sizer.Add(vsizer, 0, wx.ALL | wx.EXPAND, 0)

        self.sizer.Add(
            wx.StaticLine(self, style=wx.LI_VERTICAL),
            0,
            wx.LEFT | wx.RIGHT | wx.EXPAND,
            5,
        )

        # construct hex editor
        self.construct_hex_editor = ConstructHexEditor(
            self,
            construct=default_gallery_item.construct,
            contextkw=default_gallery_item.contextkw,
        )
        self.construct_hex_editor.construct_editor.expand_all()
        self.sizer.Add(self.construct_hex_editor, 1, wx.ALL | wx.EXPAND, 0)

        self.SetSizer(self.sizer)

        # Connect Events ##################################################
        self.gallery_selector_lbx.Bind(
            wx.EVT_LISTBOX, self.on_gallery_selection_changed
        )
        self.reload_btn.Bind(wx.EVT_BUTTON, self.on_gallery_selection_changed)
        self.clear_binary_btn.Bind(wx.EVT_BUTTON, self.on_clear_binary_clicked)
        self.example_selector_lbx.Bind(
            wx.EVT_LISTBOX, self.on_example_selection_changed
        )

        self.load_binary_file_btn.Bind(wx.EVT_BUTTON, self.on_load_binary_file_clicked)

        # Emulate Selection Click
        self.on_gallery_selection_changed(None)

    def on_gallery_selection_changed(self, event):
        selection = self.gallery_selector_lbx.GetStringSelection()
        gallery_item = self.construct_gallery[selection]
        if gallery_item is None:
            self.gallery_selector_lbx.SetSelection(
                self.gallery_selection
            )  # restore old selection
            return

        # save currently shown selection
        self.gallery_selection = self.gallery_selector_lbx.GetSelection()

        self.example_selector_lbx.Clear()
        if len(gallery_item.example_binarys) > 0:
            self.example_selector_lbx.InsertItems(
                list(gallery_item.example_binarys.keys()), 0
            )
            self.example_selector_lbx.SetStringSelection(
                list(gallery_item.example_binarys.keys())[0]
            )

            example = self.example_selector_lbx.GetStringSelection()
            example_binary = self.construct_gallery[selection].example_binarys[example]
        else:
            example_binary = bytes(0)

        self.Freeze()
        self.construct_hex_editor.binary = example_binary
        self.construct_hex_editor.construct = gallery_item.construct
        self.construct_hex_editor.contextkw = gallery_item.contextkw
        self.construct_hex_editor.refresh()
        self.construct_hex_editor.construct_editor.expand_all()
        self.Thaw()

    def on_clear_binary_clicked(self, event):
        self.example_selector_lbx.SetSelection(wx.NOT_FOUND)
        self.construct_hex_editor.binary = bytes()
        self.construct_hex_editor.refresh()

    def on_example_selection_changed(self, event):
        selection = self.gallery_selector_lbx.GetStringSelection()
        example = self.example_selector_lbx.GetStringSelection()
        example_binary = self.construct_gallery[selection].example_binarys[example]

        # Set example binary
        self.construct_hex_editor.binary = example_binary
        self.construct_hex_editor.refresh()

    def on_load_binary_file_clicked(self, event):
        with wx.FileDialog(
            self,
            "Open binary file",
            wildcard="binary files (*.*)|*.*",
            style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST,
        ) as fileDialog:

            if fileDialog.ShowModal() == wx.ID_CANCEL:
                return  # the user changed their mind

            # Proceed loading the file chosen by the user
            pathname = Path(fileDialog.GetPath())
            with open(pathname, "rb") as file:
                self.construct_hex_editor.binary = file.read()
                self.construct_hex_editor.refresh()


icon = PyEmbeddedImage(
    b"iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABGdBTUEAALGPC/xhBQAAACBj"
    b"SFJNAAB6JgAAgIQAAPoAAACA6AAAdTAAAOpgAAA6mAAAF3CculE8AAAABmJLR0QA/wD/AP+g"
    b"vaeTAAAAB3RJTUUH5QMVEwoXzWaNiAAACetJREFUWMPFl3tQVNcdxz/37oNlF3mJoAS0pviq"
    b"gqiMxRCjkGhSajMTw0wYJyEgiWmijulfRqfEvGge2nSsto5TnFjrMyaTxqmP0WiqEDWIFDS8"
    b"Nmp4s6zswr5Z9t57+geyFSHT6V89M3fOueeec37f3+P7+50L/+cmjTdZ9t6OpwOBwcUCKUKS"
    b"JCQJhLh/hRg9EqPnxb3JB3tN0zDo9T2mCOPn5ds2t40L4Nebtv41I31OUd5jjxATM4GRE4UA"
    b"gQiPRw4WCISmoaoamqahaiqaOtyr6r2xqqCqGoHBAB1dNi5X13kiI/RPfPzB29W6+4W/+9Gu"
    b"rCmTE//0xGM/Z/tHH1L3r1p+teqX1NfVUf7eu0xOSmR+RgaBgJ+3tm2jsvISK1c8Qd/du5S/"
    b"9y6qqrD00UdxDfTzwfu/w+N2k5e7HIvFzK4/7qSq8hKPZC9mVtr0iIvfVM9JTXnokHw/AEVR"
    b"5i1ZvJAdO3awYcMG5s+fz6effkp2djYrV67EarUihGDPnj3k5+ezZs0adu3aRVpaGs8//zz1"
    b"9fUIIUhOTmbdunXU1taiaRrnz5/HbDazdetWdu/ezUPJk5kyefK8yn+enzYKgNFo1JnNkbjd"
    b"bqZOncrMmTNpbGxEr9cTFxeHuBcIra2tzJs3jxkzZtDc3Iwsy0ycOBEATdMQQoTXq6pKa2sr"
    b"M2bMYMKECfh8PhRFIdIUYfD7fZP09wOQJQkhBJGRkaiqit1uJyYmBiHEqCchIYGBgQF8Pt+Y"
    b"76qqomlaGIiiKMTHxzMwMMDQ0BCKoowoIgGmUQBGAm7t2rVs3rwZr9dLWVkZN27c4MCBA3i9"
    b"XhYtWkRxcTHl5eUIISguLqazs5OdO3ditVo5ffo0CxcOu7GpqYkTJ06wbNky3njjDWpqasjL"
    b"y0On+0/ojWLB+7//c+kvViyrSJ87G4/Hg9/vZ9KkSQghsNvtCCEwGo1ER0fj9/txu90kJCQg"
    b"hKCvr49QKARATEwMTqeTYDCIEIKYmBgCgQAOh4MJEyYgSRJHvzjt//CdrU/LYw0wLOzrr78O"
    b"vzudTqqrqwmFQkRHRyOE4MKFCzQ2NiLEML/r6uq4fv16WFh9fT12u524uDiCwSC1tbX09vZi"
    b"NBpHNJcAxrhAALdu3eLYsWN4vV6ee+45Wltb+fLLL+no6GDdunUcOXIEq9WKLMu4XC50Oh3n"
    b"zp0jNTUVm83G8uXLuXbtGu3t7Wzfvh2/38/Nmzepra1l165do+TJo4UPJ5rs7Gzy8vLCgZSZ"
    b"mclTTz0VDrTKykoKCwspKiri9OnTVFVVUVBQQHFxMWfPnmXy5MkUFhYihECSJOLj4ykqKhrW"
    b"XJJGxYD+QfXFPbOPUO7+8f000+l0aJrG0NAQQggMBgNAOA6GU7iELMvIsoxOp0OSJAwGA6qq"
    b"/giAeyyw2+20trZiNBpxuVwoisLt27dxOBw4HA5ycnI4deoUg4ODZGVlkZqaypkzZ5gyZQpZ"
    b"WVmEQiGam5vp7+/HbreTmJhIQ0MDHo8Hm80WzhnjukAIgcvlIj4+HovFQltbG16vF71eT2Ji"
    b"Irdv36agoICJEycSGxvLM888Q15eHunp6SiKQklJCaqq0t3dTW5uLk1NTQDcuXOH/Px8Ghsb"
    b"h10gjUPD8u27S1fk5lRkzJszyvSapoUTzINJ6cEmSVK4f/C7pmkAeDwe/nLg2OAHb29dNdoC"
    b"9yh18uRJcnNzOXPmDJqmUVNTw5o1a1i7di29vb3U1NTw7LPPUlFRAUBXVxcvvvgipaWlfP/9"
    b"97hcLkpKSigtLaW+vp5QKMRLL73EK6+8QlVVFbIsI93TXf8gAFVVycnJob29nYGBAVRV5ZNP"
    b"PmHLli24XC4OHjzIpk2bWL9+PdevX0eSJI4fP05RURGZmZmUl5eTlZXFqlWrWL16NS+//DIF"
    b"BQVkZ2fz2muvUVhYyJ49e8aPAU0Mm7qzsxOn08ndu3fp6enB4XBgMpmIioqio6ODUCiEXq8P"
    b"W6yvr4/U1FSSkpLo6enB6XSSmpqK2WzG4/HQ29tLSkpKeM8IU8ZaQBOo9/wEMDQ0hNfrJTo6"
    b"mq6uLmw2G5MmTSIUCoX9Kcsy06dPx2azERUVRXJyMtOnT6e7uxuHw4HFYiEtLY3m5mYCgUCY"
    b"sqqmhsYCQKCpKvX19Vy8eBFJkkhJSaGoqIi9e/eiqiqvv/46ra2tHDhwAJ/Px9mzZ3nhhRco"
    b"KyvD5/Oxfv16MjIy2LJlC+fOnePVV19l6dKlfPXVV2zcuJHS0lJs9j5hbbF2A2IUC7aVf1ya"
    b"k72wIi7aEjaTTqfDZDKFs5fFYsFgMCCEQK/XYzQaMZlMAASDQSIjI8NM+K6xhQG3Z7g8qyp+"
    b"vx8kCYfDyTtvvbm3u7N9z5gg1FQVj8fDkSNHSEpKYvXq1QSDQQ4ePEhsbCwlJSUoisK+ffsw"
    b"GAxs3LiR3t5eDh06RGZmJitWrECSJL6tqWNKUiLzfjaLYDCI2+2mv7+fYDDIP06esna03blo"
    b"NlvaxqVhRUUFGRkZuFwurly5wv79+3n44YeJiIjgxIkTHD58mNjYWKZNm8a+ffswmUykpKRw"
    b"8uRJgsEg1deuEz0hip9MS0HTNILBIF6vl6GhIb759rr36MH9f9PpdI1+v29gTCYcuQnNnj2b"
    b"RYsWcePGDdra2li4cCFLliyhtraWhoYGli5dyvLly7l69SqxsbEsWLAAIQQtLS1cqa5hzqw0"
    b"FEXB7/fT39+Pz+fjh7ZOdftHHx72+7xXzZao5nGLkaZp4QISCATQ6/Xo9cPL/H5/2O8jh48U"
    b"IYDBwUF6enpwu92EQiGCwSBOpxOv10tP713tt2Vln3W1/3DGbLbUeNyu4BgAI3e5nJwcjh49"
    b"SkdHB4WFhcydO5f9+/fj8XhYtWoV8fHxVFRUoNfrefLJJ3E4HBw/fhyr1UpLSwtCiPCNyePx"
    b"0NXTq5W9+daJlsabX0SYTJf9ft/AuHkgpCiEFIX8/Hyam5uRJInZs2eHfaxpGunp6ZjNZhIS"
    b"EvD5fCxYsACdTsfKlSt5/PHHuXXrFm6/HZfLhcfj4dYP7UpZWdln1qbvvjCZIqsGBwO998sc"
    b"BUCSpLaOLhvTpz7EnDlzwjQUQjBz5kwMBkOYgmlpaRgMBvR6PbIss3jxYiRJwmg0Yutz4Xa7"
    b"uXDpsvvjHdsP9XS1n42IMF0eHAzYHyxeowAY9LpLDc236+Nio+fPnfVTZFlG0zQURUHTNEKh"
    b"UPhyMRInsiyHLx8And29IiY2ju1/2P3d3z8/dngw4K82W6Jq/T7vAOO0Mf+GG36zNf5O6533"
    b"IyIiH42KslgQQh61Thq9/f5XVVWV7u5Oe0N9baXd1vWtLOuaLFEWq8ftHuJH2rh/x0uW5uqa"
    b"Gm4kDzgdkwAj/1sTsiy7jcaIrsHBgPu/Lf4340NlvbmvI1QAAAAldEVYdGRhdGU6Y3JlYXRl"
    b"ADIwMjEtMDMtMjFUMTk6MTA6MjIrMDA6MDDMGKfHAAAAJXRFWHRkYXRlOm1vZGlmeQAyMDIx"
    b"LTAzLTIxVDE5OjEwOjIyKzAwOjAwvUUfewAAAABJRU5ErkJggg=="
)


def main():
    if sys.platform == "win32":
        # Windows Icon fix: https://stackoverflow.com/a/1552105
        import ctypes

        myappid = "timrid.construct_hex_editor"
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

    inspect = False
    if inspect is True:
        import wx.lib.mixins.inspection as wit
        app = wit.InspectableApp()
    else:
        app = wx.App(False)

    frame = ConstructGalleryFrame(None)
    frame.Show(True)
    app.MainLoop()


if __name__ == "__main__":
    main()


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


@dataclasses.dataclass
class RenamedTest(cst.DataclassMixin):
    doc: int = cst.csfield(cs.Int8sb, doc="This is a one line documentation")
    doc_multiline: int = cst.csfield(
        cs.Int8sb,
        doc="""
        This is a multiline documentation.
        Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. 
        Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.

        Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.
        """,
    )
    doc2: int = cst.csfield(cs.Int8sb * "This is a one line documentation")
    doc2_multiline: int = cst.csfield(
        cs.Int8sb
        * """
        This is a multiline documentation.
        Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. 
        Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.

        Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.
        """
    )


constr = "renamed_test" / cst.DataclassStruct(RenamedTest)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Zeros": bytes(constr.sizeof()),
    },
)


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


constr = cs.Struct(
    "signature" / cs.Bytes(23),
    "data_start_pos" / cs.Tell,
    "data_peek" / cs.Peek(cs.Array(5, cs.Int24ub)),
    cs.Seek(15, 1),
    "pos_after_seek" / cs.Tell,
    "data_pointer"
    / cs.Pointer(lambda ctx: ctx.data_start_pos, cs.Array(5, cs.Int24ub)),
)


gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Zeros": bytes(23 + 15),
        "1": b"TestPointerPeekSeekTell0123456789abcde",
    },
)


import dataclasses
import construct as cs
import typing as t


@dataclasses.dataclass
class GalleryItem:
    construct: "cs.Construct[t.Any, t.Any]"
    contextkw: t.Dict[str, t.Any] = dataclasses.field(default_factory=dict)
    example_binarys: t.Dict[str, bytes] = dataclasses.field(default_factory=dict)



import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


@dataclasses.dataclass
class Entry(cst.DataclassMixin):
    id: int = cst.csfield(cs.Int8sb)
    width: int = cst.csfield(cs.Int8sb)
    height: int = cst.csfield(cs.Int8sb)


constr = cs.GreedyRange(cst.DataclassStruct(Entry))

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "5": bytes([1, 10, 10, 2, 10, 10, 3, 18, 18, 4, 10, 10, 5, 20, 20]),
        "1": bytes([1, 10, 10]),
        "Zeros": bytes([]),
    },
)


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


@dataclasses.dataclass
class DataclassStructTest(cst.DataclassMixin):
    width: int = cst.csfield(cs.Int8sb, doc="Das hier ist die Dokumentation von 'width'")
    height: int = cst.csfield(cs.Int8sb, doc="Und hier von 'height")
    update: int = cst.csfield(cs.Int8sb)

    @dataclasses.dataclass
    class Nested(cst.DataclassMixin):
        nested_width: int = cst.csfield(cs.Int16sb)
        nested_height: int = cst.csfield(cs.Int16sb)
        nested_bytes: bytes = cst.csfield(cs.Bytes(2))
        nested_array: t.List[int] = cst.csfield(cs.Array(2, cs.Int8sb))

    nested: Nested = cst.csfield((cst.DataclassStruct(Nested)))


constr = cst.DataclassStruct(DataclassStructTest)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Zeros": bytes(constr.sizeof()),
    },
)



import construct as cs
import construct_typed as cst
import dataclasses
from . import GalleryItem


@dataclasses.dataclass
class TBitsStructTest(cst.DataclassMixin):
    @dataclasses.dataclass
    class Nested(cst.DataclassMixin):
        test_bit: int = cst.csfield(cs.Bit)
        test_nibble: int = cst.csfield(cs.Nibble)
        test_bits_1: int = cst.csfield(cs.BitsInteger(3))
        test_bits_2: int = cst.csfield(cs.BitsInteger(6))
        test_bits_3: int = cst.csfield(cs.BitsInteger(2))

    nested: Nested = cst.csfield(cs.ByteSwapped(cst.DataclassBitStruct(Nested)))

    nested_reverse: Nested = cst.csfield(cst.DataclassBitStruct(Nested, reverse=True))


constr = cst.DataclassStruct(TBitsStructTest)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Zeros": bytes(constr.sizeof()),
    },
)



import dataclasses
import enum
import string
import typing as t
from typing import Any, Dict, List, Optional, Type, Union

import arrow
import construct as cs
import construct_editor.widgets.construct_editor as construct_editor
import construct_typed as cst
import wx
import wx.adv
from construct_editor.helper.preprocessor import (
    GuiMetaData,
    IncludeGuiMetaData,
    add_gui_metadata,
    get_gui_metadata,
)


def evaluate(param, context):
    return param(context) if callable(param) else param


def int_to_str(integer_format: "construct_editor.IntegerFormat", val: int) -> str:
    if isinstance(val, str):
        return val  # tolerate string
    if integer_format is construct_editor.IntegerFormat.Hex:
        return f"0x{val:X}"
    return f"{val}"


def str_to_int(s: str) -> int:
    if len(s) == 0:
        s = "0"

    # convert string to int
    # (base=0 means, that eg. 0x, 0b prefixes are allowed)
    i = int(s, base=0)

    return i


def str_to_bytes(s: str) -> bytes:
    return bytes.fromhex(s)

# #####################################################################################################################
# GUI Elements ########################################################################################################
# #####################################################################################################################
class ObjPanel(wx.Panel):
    """ Base class for a panel that shows the value and allows modifications of it. """

    def get_new_obj(self) -> Any:
        raise NotImplementedError()


class ObjPanel_Empty(ObjPanel):
    def __init__(self, parent):
        super().__init__(parent)

        # Obj
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.obj_txt = wx.TextCtrl(
            self,
            wx.ID_ANY,
            wx.EmptyString,
            wx.DefaultPosition,
            wx.Size(-1, -1),
            wx.TE_READONLY,
        )
        hsizer.Add(self.obj_txt, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 0)

        self.SetSizer(hsizer)
        self.Layout()


class ObjPanel_Default(ObjPanel):
    def __init__(self, parent, entry: "EntryConstruct"):
        super().__init__(parent)
        self.entry = entry

        # Test if the obj of the entry is available
        if self.entry.obj is None:
            return

        # Obj
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.obj_txt = wx.TextCtrl(
            self,
            wx.ID_ANY,
            self.entry.obj_str,
            wx.DefaultPosition,
            wx.Size(-1, -1),
            wx.TE_READONLY,
        )
        hsizer.Add(self.obj_txt, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 0)

        self.SetSizer(hsizer)
        self.Layout()

    def get_new_obj(self) -> Any:
        return self.entry.obj


class ObjPanel_String(ObjPanel):
    def __init__(self, parent, entry: "EntryConstruct"):
        super().__init__(parent)
        self.entry = entry

        # Test if the obj of the entry is available
        if self.entry.obj is None:
            return

        # Obj
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.obj_txt = wx.TextCtrl(
            self,
            wx.ID_ANY,
            self.entry.obj_str,
        )
        hsizer.Add(self.obj_txt, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 0)

        self.SetSizer(hsizer)
        self.Layout()

    def get_new_obj(self) -> Any:
        val_str: str = self.obj_txt.GetValue()
        return val_str


class ObjPanel_Integer(ObjPanel):
    def __init__(self, parent, entry: "EntryConstruct"):
        super().__init__(parent)
        self.entry = entry

        # Test if the obj of the entry is available
        if self.entry.obj is None:
            return

        # Obj
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.obj_txtctrl = wx.TextCtrl(self, wx.ID_ANY, self.entry.obj_str)
        hsizer.Add(self.obj_txtctrl, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 0)

        self.SetSizer(hsizer)
        self.Layout()

    def get_new_obj(self) -> Any:
        val_str: str = self.obj_txtctrl.GetValue()

        try:
            # convert string to integer
            new_obj = str_to_int(val_str)
        except Exception:
            new_obj = val_str  # this will probably result in a building error

        return new_obj


class ObjPanel_Bytes(ObjPanel):
    def __init__(self, parent, entry: "EntryConstruct"):
        super().__init__(parent)
        self.entry = entry

        # Test if the obj of the entry is available
        if self.entry.obj is None:
            return

        # Obj
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.obj_txtctrl = wx.TextCtrl(self, wx.ID_ANY, self.entry.obj_str)
        hsizer.Add(self.obj_txtctrl, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 0)

        self.SetSizer(hsizer)
        self.Layout()

    def get_new_obj(self) -> Any:
        val_str: str = self.obj_txtctrl.GetValue()

        try:
            # convert string to bytes
            new_obj = str_to_bytes(val_str)
        except Exception:
            new_obj = val_str  # this will probably result in a building error

        return new_obj


@dataclasses.dataclass
class EnumItem:
    name: str
    value: int


class ObjPanel_Enum(ObjPanel):
    def __init__(self, parent, entry: Union["EntryTEnum", "EntryEnum"]):
        super().__init__(parent)
        self.entry = entry

        # Test if the obj of the entry is available
        if self.entry.obj is None:
            return

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        self.obj_combobox = wx.ComboBox(
            self,
            style=wx.CB_DROPDOWN,
        )
        hsizer.Add(self.obj_combobox, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 0)

        items = self.entry.get_enum_items()
        for pos, item in enumerate(items):
            self.obj_combobox.Insert(
                item=f"{int_to_str(entry.model.integer_format, item.value)} ({item.name})",
                pos=pos,
                clientData=item,
            )
        item = self.entry.get_enum_item_from_obj()
        sel_item_str = (
            f"{int_to_str(entry.model.integer_format, item.value)} ({item.name})"
        )
        self.obj_combobox.SetStringSelection(sel_item_str)
        self.obj_combobox.SetValue(sel_item_str)  # show even if it is not in the list

        self.SetSizer(hsizer)
        self.Layout()

    def get_new_obj(self) -> Any:
        val_str: str = self.obj_combobox.GetValue()
        if len(val_str) == 0:
            val_str = "0"

        val_str = val_str.split()[0]
        new_obj = self.entry.conv_str_to_obj(val_str)
        return new_obj


@dataclasses.dataclass
class FlagsEnumItem:
    name: str
    value: int
    checked: bool


class FlagsEnumComboPopup(wx.ComboPopup):
    def __init__(
        self,
        combo_ctrl: wx.ComboCtrl,
        entry: Union["EntryTFlagsEnum", "EntryFlagsEnum"],
    ):
        super().__init__()
        self.combo_ctrl = combo_ctrl
        self.entry = entry
        self.clbx: Optional[wx.CheckListBox] = None

    def on_motion(self, evt):
        item = self.clbx.HitTest(evt.GetPosition())
        if item != wx.NOT_FOUND:
            # only select if not selected prevents flickering
            if not self.clbx.IsSelected(item):
                self.clbx.Select(item)

    def on_left_down(self, evt):
        item = self.clbx.HitTest(evt.GetPosition())
        if item != wx.NOT_FOUND:
            # select the new item in the gui
            items = list(self.clbx.GetCheckedItems())
            if item in items:
                items.remove(item)
            else:
                items.append(item)
            self.clbx.SetCheckedItems(items)

            # refresh shown string
            self.combo_ctrl.SetValue(self.GetStringValue())

    def Create(self, parent):
        self.clbx = wx.CheckListBox(parent)
        self.clbx.Bind(wx.EVT_MOTION, self.on_motion)
        self.clbx.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        return True

    # Return the widget that is to be used for the popup
    def GetControl(self):
        return self.clbx

    # Return final size of popup. Called on every popup, just prior to OnPopup.
    # minWidth = preferred minimum width for window
    # prefHeight = preferred height. Only applies if > 0,
    # maxHeight = max height for window, as limited by screen size
    #   and should only be rounded down, if necessary.
    def GetAdjustedSize(self, minWidth, prefHeight, maxHeight):
        row_height = self.clbx.GetCharHeight() + 2
        row_count = self.clbx.GetCount()
        prefHeight = min(row_height * row_count + 4, prefHeight)
        return wx.ComboPopup.GetAdjustedSize(self, minWidth, prefHeight, maxHeight)

    def get_flagsenum_items(self) -> t.List["FlagsEnumItem"]:
        # read all flagsenum items and modify checked status
        flagsenum_items: t.List[FlagsEnumItem] = []
        for item in range(self.clbx.GetCount()):
            flagsenum_item: FlagsEnumItem = self.clbx.GetClientData(item)
            flagsenum_item.checked = self.clbx.IsChecked(item)
            flagsenum_items.append(flagsenum_item)
        return flagsenum_items

    def GetStringValue(self):
        flagsenum_items = self.get_flagsenum_items()
        temp_obj = self.entry.conv_flagsenum_items_to_obj(flagsenum_items)
        return self.entry.conv_obj_to_str(temp_obj)


class ObjPanel_FlagsEnum(ObjPanel):
    def __init__(self, parent, entry: Union["EntryTFlagsEnum", "EntryFlagsEnum"]):
        super().__init__(parent)
        self.entry = entry

        obj = self.entry.obj

        # Test if the obj of the entry is available
        if obj is None:
            return

        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        self.combo_ctrl = wx.ComboCtrl(self, style=wx.CB_READONLY)
        self.popup_ctrl = FlagsEnumComboPopup(self.combo_ctrl, self.entry)
        self.combo_ctrl.SetPopupControl(self.popup_ctrl)

        # Initialize CheckListBox
        items = self.entry.get_flagsenum_items_from_obj()
        for pos, item in enumerate(items):
            self.popup_ctrl.clbx.Insert(
                item=f"{int_to_str(entry.model.integer_format, item.value)} ({item.name})",
                pos=pos,
                clientData=item,
            )
            self.popup_ctrl.clbx.Check(pos, item.checked)

        self.combo_ctrl.SetValue(self.popup_ctrl.GetStringValue())

        hsizer.Add(self.combo_ctrl, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 0)

        self.SetSizer(hsizer)
        self.Layout()

    def get_new_obj(self) -> Any:
        flagsenum_items = self.popup_ctrl.get_flagsenum_items()
        new_obj = self.entry.conv_flagsenum_items_to_obj(flagsenum_items)
        return new_obj


class ObjPanel_Timestamp(ObjPanel):
    def __init__(self, parent, entry: "EntryTimestamp"):
        super().__init__(parent)
        self.entry = entry

        # Test if the obj of the entry is available
        if entry.obj is None:
            return
        if not isinstance(entry.obj, arrow.Arrow):
            return

        # Obj
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        dt = self.entry.obj.datetime
        wx_datetime = wx.DateTime(
            day=dt.day,
            month=dt.month - 1,  # in wx.adc.DatePickerCtrl the month start with 0
            year=dt.year,
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            millisec=dt.microsecond // 1000,
        )

        self.date_picker = wx.adv.DatePickerCtrl(
            self,
            wx.ID_ANY,
            wx_datetime,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.adv.DP_DROPDOWN | wx.adv.DP_SHOWCENTURY,
        )
        hsizer.Add(self.date_picker, 0, wx.LEFT, 0)

        self.time_picker = wx.adv.TimePickerCtrl(
            self,
            wx.ID_ANY,
            wx_datetime,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.adv.TP_DEFAULT,
        )
        hsizer.Add(self.time_picker, 0, wx.LEFT, 5)

        self.obj_txtctrl = wx.TextCtrl(
            self,
            wx.ID_ANY,
            self.entry.obj_str,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.TE_READONLY,
        )
        hsizer.Add(self.obj_txtctrl, 1, wx.LEFT | wx.ALIGN_CENTER_VERTICAL, 5)

        self.SetSizer(hsizer)
        self.Layout()

    def get_new_obj(self) -> Any:
        date: wx.DateTime = self.date_picker.GetValue()
        time: wx.DateTime = self.time_picker.GetValue()
        new_obj = arrow.Arrow(
            year=date.year,
            month=date.month + 1,  # in wx.adc.DatePickerCtrl the month start with 0
            day=date.day,
            hour=time.hour,
            minute=time.minute,
            second=time.second,
        )
        return new_obj


# #####################################################################################################################
# DVC Entries #########################################################################################################
# #####################################################################################################################

# EntryConstruct ######################################################################################################
class EntryConstruct(object):
    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Construct[Any, Any]",
    ):
        self.model = model
        self._parent = parent
        self._construct = construct

        # This is set from the model, when the dvc item for this entry is created.
        self._dvc_item = None
        self._dvc_item_expanded = False

    # default "parent" ########################################################
    @property
    def parent(self) -> Optional["EntryConstruct"]:
        return self._parent

    # default "construct" #####################################################
    @property
    def construct(self) -> "cs.Construct[Any, Any]":
        return self._construct

    # default "obj" ###########################################################
    @property
    def obj(self) -> Any:
        path = self.path
        obj = self.model.root_obj
        for p in path[1:]:
            if isinstance(obj, dict) or isinstance(obj, cst.DataclassMixin):
                obj = obj[p]
            elif isinstance(obj, list):
                obj = obj[int(p)]
        return obj

    @obj.setter
    def obj(self, val: Any):
        path = self.path
        obj = self.model.root_obj
        for p in path[1:-1]:
            if isinstance(obj, dict) or isinstance(obj, cst.DataclassMixin):
                obj = obj[p]
            elif isinstance(obj, list):
                obj = obj[int(p)]

        if isinstance(obj, dict) or isinstance(obj, cst.DataclassMixin):
            obj[path[-1]] = val
        elif isinstance(obj, list):
            obj[int(path[-1])] = val

    # default "obj_str" #######################################################
    @property
    def obj_str(self) -> str:
        return str(self.obj)

    # default "obj_metadata" ##################################################
    @property
    def obj_metadata(self) -> t.Optional[GuiMetaData]:
        return get_gui_metadata(self.obj)

    # default "name" ##########################################################
    @property
    def name(self) -> str:
        if self.construct.name is not None:
            return self.construct.name
        else:
            return ""

    # default "docs" ##########################################################
    @property
    def docs(self) -> str:
        return self.construct.docs

    # default "typ_str" #######################################################
    @property
    def typ_str(self) -> str:
        return repr(self.construct)

    # default "subentries" ####################################################
    @property
    def subentries(self) -> Optional[List["EntryConstruct"]]:
        return None

    # default "dvc_item" ######################################################
    @property
    def dvc_item(self) -> Any:
        return self._dvc_item

    @dvc_item.setter
    def dvc_item(self, val: Any):
        self._dvc_item = val

    # default "dvc_item_expanded" #############################################
    @property
    def dvc_item_expanded(self) -> bool:
        return self._dvc_item_expanded

    @dvc_item_expanded.setter
    def dvc_item_expanded(self, val: bool):
        self._dvc_item_expanded = val

    # default "dvc_item_expanded" #############################################
    def dvc_item_restore_expansion(self):
        """ Restore the expansion state, recursively """
        dvc_item = self.dvc_item
        if dvc_item is not None:
            if self.dvc_item_expanded:
                self.model.dvc.Expand(self.dvc_item)
            else:
                self.model.dvc.Collapse(self.dvc_item)
        subentries = self.subentries
        if subentries is not None:
            for subentry in subentries:
                subentry.dvc_item_restore_expansion()

    # default "add_nonstruct_subentries_to_list" ##############################
    def create_flat_subentry_list(self, flat_subentry_list: List["EntryConstruct"]):
        """ Create a flat list with all subentires, recursively """
        subentries = self.subentries
        if subentries is not None:
            for subentry in subentries:
                subentry.create_flat_subentry_list(flat_subentry_list)
        else:
            flat_subentry_list.append(self)

    # default "create_obj_panel" ##############################################
    def create_obj_panel(self, parent) -> ObjPanel:
        """ This method is called, when the user clicks an entry """
        return ObjPanel_Default(parent, self)

    # default "modify_context_menu" ###########################################
    def modify_context_menu(self, menu: "construct_editor.ContextMenu"):
        """ This method is called, when the user right clicks an entry and a ContextMenu is created """
        pass

    # default "path" ##########################################################
    @property
    def path(self) -> List[str]:
        parent = self.parent
        if parent is not None:
            path = parent.path
        else:
            path = []

        # Append name if available
        name = self.name
        if name != "":
            path.append(name)

        return path


# EntrySubconstruct ###################################################################################################
class EntrySubconstruct(EntryConstruct):
    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Subconstruct[Any, Any, Any, Any]",
    ):
        super().__init__(model, parent, construct)

        self.subentry = create_entry_from_construct(model, self, construct.subcon)

    # pass throught "obj_str" to subentry #####################################
    @property
    def obj_str(self) -> Any:
        return self.subentry.obj_str

    # pass throught "typ_str" to subentry #####################################
    @property
    def typ_str(self) -> str:
        return self.subentry.typ_str

    # pass throught "subentries" to subentry ##################################
    @property
    def subentries(self) -> Optional[List["EntryConstruct"]]:
        return self.subentry.subentries

    # pass throught "dvc_item" to subentry ####################################
    @property
    def dvc_item(self) -> Any:
        return self.subentry.dvc_item

    @dvc_item.setter
    def dvc_item(self, val: Any):
        self.subentry.dvc_item = val

    # pass throught "dvc_item_expanded" to subentry ###########################
    @property
    def dvc_item_expanded(self) -> bool:
        return self.subentry.dvc_item_expanded

    @dvc_item_expanded.setter
    def dvc_item_expanded(self, val: bool):
        self.subentry.dvc_item_expanded = val

    # pass throught "create_obj_panel" to subentry ############################
    def create_obj_panel(self, parent) -> ObjPanel:
        return self.subentry.create_obj_panel(parent)

    # pass throught "modify_context_menu" to subentry #########################
    def modify_context_menu(self, menu: "construct_editor.ContextMenu"):
        return self.subentry.modify_context_menu(menu)


# EntryAdapter ########################################################################################################
class AdapterPanelType(enum.Enum):
    Default = enum.auto()
    Integer = enum.auto()
    String = enum.auto()


@dataclasses.dataclass
class AdapterMapping:
    name: str
    panel: AdapterPanelType


@dataclasses.dataclass
class AdapterInstanceMapping(AdapterMapping):
    """ Mapping for a Adapter Instance (normally a cs.ExprAdapter) """

    adapter_instance: "cs.Adapter[Any, Any, Any, Any]"


@dataclasses.dataclass
class AdapterClassMapping(AdapterMapping):
    """ Mapping for a Adapter Instance (normally a cs.Adapter) """

    adapter_class: Type["cs.Adapter[Any, Any, Any, Any]"]


def add_adapter_mapping(
    adapter_mapping: t.Union[AdapterInstanceMapping, AdapterClassMapping]
):
    """ Add a Mapping for a custom adapter construct """

    class EntryAdapter(EntryConstruct):
        def __init__(
            self,
            model: "construct_editor.ConstructEditorModel",
            parent: Optional["EntryConstruct"],
            construct: "cs.Subconstruct[Any, Any, Any, Any]",
        ):
            super().__init__(model, parent, construct)

        @property
        def typ_str(self) -> str:
            return adapter_mapping.name

        @property
        def obj_str(self) -> Any:
            return str(self.obj)

        def create_obj_panel(self, parent) -> ObjPanel:
            if adapter_mapping.panel == AdapterPanelType.Integer:
                return ObjPanel_Integer(parent, self)
            elif adapter_mapping.panel == AdapterPanelType.String:
                return ObjPanel_String(parent, self)
            else:
                return ObjPanel_Default(parent, self)

    if isinstance(adapter_mapping, AdapterInstanceMapping):
        mapping = SigletonEntryMapping(adapter_mapping.adapter_instance, EntryAdapter)
    elif isinstance(adapter_mapping, AdapterClassMapping):
        mapping = ClassEntryMapping(adapter_mapping.adapter_class, EntryAdapter)
    else:
        raise ValueError(
            "adapter_mapping has to be of type 'AdapterInstanceMapping' or 'AdapterClassMapping'"
        )

    construct_entry_mapping.append(mapping)


# EntryStruct #########################################################################################################
class EntryStruct(EntryConstruct):
    construct: "cs.Struct[Any, Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Struct[Any, Any]",
    ):
        super().__init__(model, parent, construct)

        # change default row infos
        self._subentries = []

        # create sub entries
        for subcon in self.construct.subcons:
            subentry = create_entry_from_construct(model, self, subcon)
            self._subentries.append(subentry)

    @property
    def subentries(self) -> Optional[List["EntryConstruct"]]:
        return self._subentries

    @property
    def typ_str(self) -> str:
        return "Struct"

    @property
    def obj_str(self) -> str:
        return ""

    def create_obj_panel(self, parent) -> ObjPanel:
        return ObjPanel_Default(parent, self)  # TODO: create panel for cs.Struct


# EntryArray ##########################################################################################################
class EntryArray(EntrySubconstruct):
    construct: t.Union["cs.Array[Any, Any, Any, Any]", "cs.GreedyRange[Any, Any, Any, Any]"]

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: t.Union["cs.Array[Any, Any, Any, Any]", "cs.GreedyRange[Any, Any, Any, Any]"],
    ):
        super().__init__(model, parent, construct)

        self._subentries = []

    @property
    def subentries(self) -> Optional[List["EntryConstruct"]]:
        # get length of array
        try:
            array_len = len(self.obj)
        except Exception:
            if isinstance(self.construct, cs.Array) and isinstance(self.construct.count, int):
                array_len = self.construct.count
            else:
                array_len = 1

        # append entries if not appended yet
        if len(self._subentries) != array_len:
            self._subentries.clear()
            for index in range(0, array_len):
                self.insert_entry(index)

        return self._subentries

    def insert_entry(self, index: int):
        subentry = create_entry_from_construct(
            self.model,
            self,
            cs.Renamed(self.construct.subcon, newname=str(index)),
        )
        self._subentries.append(subentry)

    @property
    def typ_str(self) -> str:
        obj = self.obj

        if isinstance(self.construct, cs.Array):
            try:
                metadata = get_gui_metadata(obj)
                count = evaluate(self.construct.count, metadata.context)
                return f"Array[{count}]"
            except Exception:
                return f"Array[{self.construct.count}]"
        else:
            try:
                return f"Array[{len(self.obj)}]"
            except Exception:
                return f"GreedyRange"

    @property
    def obj_str(self) -> str:
        return ""

    def create_obj_panel(self, parent) -> ObjPanel:
        return ObjPanel_Default(parent, self)  # TODO: create panel for cs.Array

    def modify_context_menu(self, menu: "construct_editor.ContextMenu"):
        # If the subentry has no subentries itself, it makes no sense to create a list view.
        temp_subentry = create_entry_from_construct(
            self.model, self, self.construct.subcon
        )
        if temp_subentry.subentries is None:
            return

        menu.Append(wx.MenuItem(menu, wx.ID_ANY, kind=wx.ITEM_SEPARATOR))

        def on_menu_item_clicked(event: wx.MenuEvent):
            if self in self.model.list_viewed_entries:
                self.model.list_viewed_entries.remove(self)
            else:
                self.model.list_viewed_entries.append(self)
            self.dvc_item_expanded = True
            menu.parent.reload()

        menu_item = wx.MenuItem(menu, wx.ID_ANY, "Enable List View", kind=wx.ITEM_CHECK)
        menu.Append(menu_item)
        menu_item.Check(self in self.model.list_viewed_entries)
        menu.Bind(wx.EVT_MENU, on_menu_item_clicked, menu_item)


# EntryIfThenElse #####################################################################################################
class EntryIfThenElse(EntryConstruct):
    construct: "cs.IfThenElse[Any, Any, Any, Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.IfThenElse[Any, Any, Any, Any]",
    ):
        super().__init__(model, parent, construct)

        self._subentry_then = EntryRenamed(
            self.model,
            self,
            cs.Renamed(
                self.construct.thensubcon, newname=f"If {self.construct.condfunc} then"
            ),
            exclude_from_path=True,
        )
        self._subentry_else = EntryRenamed(
            self.model,
            self,
            cs.Renamed(self.construct.elsesubcon, newname=f"Else"),
            exclude_from_path=True,
        )

        # change default row infos
        self._subentries: List[EntryConstruct] = [
            self._subentry_then,
            self._subentry_else,
        ]

    def _get_subentry(self) -> "Optional[EntryConstruct]":
        """ Evaluate the conditional function to detect the type of the subentry """
        obj = self.obj
        if obj is None:
            return None
        else:
            metadata = get_gui_metadata(obj)
            cond = evaluate(self.construct.condfunc, metadata.context)
            if cond:
                return self._subentry_then.subentry
            else:
                return self._subentry_else.subentry

    @property
    def obj_str(self) -> str:
        subentry = self._get_subentry()
        if subentry is None:
            return ""
        else:
            return subentry.obj_str

    @property
    def typ_str(self) -> str:
        subentry = self._get_subentry()
        if subentry is None:
            return "IfThenElse"
        else:
            return subentry.typ_str

    @property
    def dvc_item(self) -> Any:
        subentry = self._get_subentry()
        if subentry is None:
            return self._dvc_item
        else:
            return subentry.dvc_item

    @dvc_item.setter
    def dvc_item(self, val: Any):
        subentry = self._get_subentry()
        if subentry is None:
            self._dvc_item = val
        else:
            subentry.dvc_item = val

    @property
    def dvc_item_expanded(self) -> bool:
        subentry = self._get_subentry()
        if subentry is None:
            return self._dvc_item_expanded
        else:
            return subentry.dvc_item_expanded

    @dvc_item_expanded.setter
    def dvc_item_expanded(self, val: bool):
        subentry = self._get_subentry()
        if subentry is None:
            self._dvc_item_expanded = val
        else:
            subentry.dvc_item_expanded = val

    @property
    def subentries(self) -> Optional[List["EntryConstruct"]]:
        subentry = self._get_subentry()
        if subentry is None:
            return self._subentries
        else:
            return subentry.subentries

    def create_obj_panel(self, parent) -> ObjPanel:
        subentry = self._get_subentry()
        if subentry is None:
            return ObjPanel_Default(parent, self)
        else:
            return subentry.create_obj_panel(parent)


# EntrySwitch #########################################################################################################
class EntrySwitch(EntryConstruct):
    construct: "cs.Switch[Any, Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Switch[Any, Any]",
    ):
        super().__init__(model, parent, construct)

        self._subentries: List[EntryConstruct] = []
        self._subentry_cases: Dict[str, EntryConstruct] = {}
        self._subentry_default: Optional[EntryConstruct] = None

        for key, value in self.construct.cases.items():
            subentry_case = EntryRenamed(
                self.model,
                self,
                cs.Renamed(
                    value, newname=f"Case {self.construct.keyfunc} == {str(key)}"
                ),
                exclude_from_path=True,
            )
            self._subentry_cases[key] = subentry_case
            self._subentries.append(subentry_case)

        if self.construct.default is not None:
            self._subentry_default = EntryRenamed(
                self.model,
                self,
                cs.Renamed(self.construct.default, newname=f"Default"),
                exclude_from_path=True,
            )
            self._subentries.append(self._subentry_default)

    def _get_subentry(self) -> "Optional[EntryConstruct]":
        """ Evaluate the conditional function to detect the type of the subentry """
        obj = self.obj
        if obj is None:
            return None
        else:
            metadata = get_gui_metadata(obj)
            key = evaluate(self.construct.keyfunc, metadata.context)
            if key in self._subentry_cases:
                return self._subentry_cases[key]
            else:
                return self._subentry_default

    @property
    def obj_str(self) -> str:
        subentry = self._get_subentry()
        if subentry is None:
            return ""
        else:
            return subentry.obj_str

    @property
    def typ_str(self) -> str:
        subentry = self._get_subentry()
        if subentry is None:
            return "Switch"
        else:
            return subentry.typ_str

    @property
    def dvc_item(self) -> Any:
        subentry = self._get_subentry()
        if subentry is None:
            return self._dvc_item
        else:
            return subentry.dvc_item

    @dvc_item.setter
    def dvc_item(self, val: Any):
        subentry = self._get_subentry()
        if subentry is None:
            self._dvc_item = val
        else:
            subentry.dvc_item = val

    @property
    def dvc_item_expanded(self) -> bool:
        subentry = self._get_subentry()
        if subentry is None:
            return self._dvc_item_expanded
        else:
            return subentry.dvc_item_expanded

    @dvc_item_expanded.setter
    def dvc_item_expanded(self, val: bool):
        subentry = self._get_subentry()
        if subentry is None:
            self._dvc_item_expanded = val
        else:
            subentry.dvc_item_expanded = val

    @property
    def subentries(self) -> Optional[List["EntryConstruct"]]:
        subentry = self._get_subentry()
        if subentry is None:
            return self._subentries
        else:
            return subentry.subentries

    def create_obj_panel(self, parent) -> ObjPanel:
        subentry = self._get_subentry()
        if subentry is None:
            return ObjPanel_Default(parent, self)
        else:
            return subentry.create_obj_panel(parent)


# EntryFormatField ####################################################################################################
class EntryFormatField(EntryConstruct):
    construct: "cs.FormatField[Any, Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.FormatField[Any, Any]",
    ):
        super().__init__(model, parent, construct)

        type_mapping = {
            ">B": ("Int8ub", int, 8, False),
            ">H": ("Int16ub", int, 16, False),
            ">L": ("Int32ub", int, 32, False),
            ">Q": ("Int64ub", int, 64, False),
            ">b": ("Int8sb", int, 8, True),
            ">h": ("Int16sb", int, 16, True),
            ">l": ("Int32sb", int, 32, True),
            ">q": ("Int64sb", int, 64, True),
            "<B": ("Int8ul", int, 8, False),
            "<H": ("Int16ul", int, 16, False),
            "<L": ("Int32ul", int, 32, False),
            "<Q": ("Int64ul", int, 64, False),
            "<b": ("Int8sl", int, 8, True),
            "<h": ("Int16sl", int, 16, True),
            "<l": ("Int32sl", int, 32, True),
            "<q": ("Int64sl", int, 64, True),
            "=B": ("Int8un", int, 8, False),
            "=H": ("Int16un", int, 16, False),
            "=L": ("Int32un", int, 32, False),
            "=Q": ("Int64un", int, 64, False),
            "=b": ("Int8sn", int, 8, True),
            "=h": ("Int16sn", int, 16, True),
            "=l": ("Int32sn", int, 32, True),
            "=q": ("Int64sn", int, 64, True),
            ">e": ("Float16b", float),
            "<e": ("Float16l", float),
            "=e": ("Float16n", float),
            ">f": ("Float32b", float),
            "<f": ("Float32l", float),
            "=f": ("Float32n", float),
            ">d": ("Float64b", float),
            "<d": ("Float64l", float),
            "=d": ("Float64n", float),
        }

        # change default row infos
        self.type_infos = None
        if construct.fmtstr in type_mapping:
            self.type_infos = type_mapping[construct.fmtstr]

    def create_obj_panel(self, parent) -> ObjPanel:
        if self.type_infos[1] is int:
            return ObjPanel_Integer(parent, self)
        else:
            return ObjPanel_Default(parent, self)  # TODO: float

    @property
    def obj_str(self) -> str:
        obj = self.obj
        if (obj is None) or (self.type_infos[1] is not int):
            return str(obj)
        else:
            return int_to_str(self.model.integer_format, obj)

    @property
    def typ_str(self) -> str:
        if self.type_infos is not None:
            return self.type_infos[0]
        else:
            return "FormatField[{}]".format(repr(self.construct.fmtstr))


# EntryBytesInteger ###################################################################################################
class EntryBytesInteger(EntryConstruct):
    construct: "cs.BytesInteger[Any, Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.BytesInteger[Any, Any]",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        if self.construct.length == 3:
            if self.construct.signed is False:
                if self.construct.swapped is False:
                    return "Int24ub"
                else:
                    return "Int24ul"
            else:
                if self.construct.swapped is False:
                    return "Int24sb"
                else:
                    return "Int24sl"
        else:
            return repr(self.construct)

    @property
    def obj_str(self) -> str:
        obj = self.obj
        if obj is None:
            return str(obj)
        else:
            return int_to_str(self.model.integer_format, obj)

    def create_obj_panel(self, parent) -> ObjPanel:
        if isinstance(self.construct.length, int):
            return ObjPanel_Integer(parent, self)
        else:
            return ObjPanel_Default(parent, self)


# EntryBitsInteger ####################################################################################################
class EntryBitsInteger(EntryConstruct):
    construct: "cs.BitsInteger[Any, Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.BitsInteger[Any, Any]",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        # change default row infos
        return "BitsInteger[{}{}]".format(
            repr(self.construct.length),
            ", signed" if self.construct.signed is True else "",
        )

    @property
    def obj_str(self) -> str:
        obj = self.obj
        if obj is None:
            return str(obj)
        else:
            return int_to_str(self.model.integer_format, obj)

    def create_obj_panel(self, parent) -> ObjPanel:
        if isinstance(self.construct.length, int):
            return ObjPanel_Integer(parent, self)
        else:
            return ObjPanel_Default(parent, self)


# EntryBytes ##########################################################################################################
class EntryBytes(EntryConstruct):
    construct: t.Union["cs.Bytes[Any, Any]", "cs.Construct[bytes, bytes]"]

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: t.Union["cs.Bytes[Any, Any]", "cs.Construct[bytes, bytes]"],
    ):
        super().__init__(model, parent, construct)
        self.ascii_view = False

    @property
    def obj_str(self) -> str:
        try:
            if self.ascii_view:
                chars = []
                for b in self.obj:
                    char = chr(b)
                    if char not in string.printable:
                        char = "."
                    chars.append(char)
                return "".join(chars)
            else:
                return self.obj.hex(" ")
        except Exception:
            return str(self.obj)

    @property
    def typ_str(self) -> str:
        obj = self.obj

        if isinstance(self.construct, cs.Bytes):
            try:
                metadata = get_gui_metadata(obj)
                length = evaluate(self.construct.length, metadata.context)
                return f"Byte[{length}]"
            except Exception:
                return f"Byte[{self.construct.length}]"
        else:
            try:
                return f"Byte[{len(self.obj)}]"
            except Exception:
                return "GreedyBytes"

    def create_obj_panel(self, parent) -> ObjPanel:
        return ObjPanel_Bytes(parent, self)

    def modify_context_menu(self, menu: "construct_editor.ContextMenu"):
        menu.Append(wx.MenuItem(menu, wx.ID_ANY, kind=wx.ITEM_SEPARATOR))

        def on_menu_item_clicked(event: wx.MenuEvent):
            self.ascii_view = not self.ascii_view
            menu.parent.reload()

        menu_item = wx.MenuItem(menu, wx.ID_ANY, "ASCII View", kind=wx.ITEM_CHECK)
        menu.Append(menu_item)
        menu_item.Check(self.ascii_view)
        menu.Bind(wx.EVT_MENU, on_menu_item_clicked, menu_item)


# EntryRenamed ########################################################################################################
class EntryRenamed(EntrySubconstruct):
    construct: "cs.Renamed[Any, Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Renamed[Any, Any]",
        exclude_from_path: bool = False,
    ):
        super().__init__(model, parent, construct)
        self._exclude_from_path = exclude_from_path

        # This is for the case, when nesting Renamed like this: `Renamed(Renamed(Int8sb, newname=...), newdocs=...)`
        # because both Renamed elements have the same name, but we only need it one time. So all other cases
        # are excluded.
        if self.construct.name == self.construct.subcon.name:
            self._exclude_from_path = True

    # modified "path" #########################################################
    @property
    def path(self) -> List[str]:
        if self._exclude_from_path:
            if self.parent is not None:
                return self.parent.path
            else:
                return []
        else:
            return super().path


# EntryTell ###########################################################################################################
class EntryTell(EntryConstruct):
    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Construct[Any, Any]",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        return "Tell"


# EntrySeek ###########################################################################################################
class EntrySeek(EntryConstruct):
    construct: "cs.Seek"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Seek",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        return "Seek"

    @property
    def obj_str(self) -> str:
        return ""


# EntryPass ###########################################################################################################
class EntryPass(EntryConstruct):
    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Construct[None, None]",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        return "Pass"

    @property
    def obj_str(self) -> str:
        return ""


# EntryComputed ###########################################################################################################
class EntryComputed(EntryConstruct):
    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Computed[Any, Any]",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        return "Computed"

    @property
    def obj_str(self) -> str:
        if isinstance(self.obj, (bytes, bytearray, memoryview)):
            return self.obj.hex(" ")
        else:
            return str(self.obj)


# EntryTimestamp ###########################################################################################################
class EntryTimestamp(EntrySubconstruct):
    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.TimestampAdapter[Any, Any]",
    ):
        super().__init__(model, parent, construct)

    @property
    def obj_str(self) -> str:
        if isinstance(self.obj, (arrow.Arrow)):
            return self.obj.format("YYYY-MM-DD HH:mm:ss ZZ")
        else:
            return str(self.obj)

    def create_obj_panel(self, parent) -> ObjPanel:
        return ObjPanel_Timestamp(parent, self)


# EntryTransparentSubcon ##############################################################################################
class EntryTransparentSubcon(EntrySubconstruct):
    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Subconstruct[Any, Any, Any, Any]",
    ):
        super().__init__(model, parent, construct)


# EntryPeek ###########################################################################################################
class EntryPeek(EntrySubconstruct):
    construct: "cs.Peek"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Peek",
    ):
        super().__init__(model, parent, construct)


# EntryRawCopy #######################################################################################################
class EntryRawCopy(EntrySubconstruct):
    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.RawCopy[Any, Any, Any, Any]",
    ):
        super().__init__(model, parent, construct)

        # change default row infos


# EntryDataclassStruct ################################################################################################
class EntryDataclassStruct(EntrySubconstruct):
    construct: "cst.DataclassStruct[Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cst.DataclassStruct[Any]",
    ):
        super().__init__(model, parent, construct)

    @property
    def subentries(self) -> Optional[List["EntryConstruct"]]:
        subentries = super().subentries
        if (subentries is not None) and self.construct.reverse:
            return list(reversed(subentries))
        return subentries

    @property
    def typ_str(self) -> str:
        return self.construct.dc_type.__name__


# EntryEnum ###########################################################################################################
class EntryEnum(EntrySubconstruct):
    construct: "cs.Enum"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.Enum",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        return super().typ_str + " as Enum"

    @property
    def obj_str(self) -> str:
        try:
            return f"{int_to_str(self.model.integer_format, int(self.obj))} ({str(self.obj)})"
        except Exception:
            return str(self.obj)

    def create_obj_panel(self, parent) -> ObjPanel:
        return ObjPanel_Enum(parent, self)

    def get_enum_items(self) -> t.List[EnumItem]:
        """ Get items to show in the ComboBox """
        items: t.List[EnumItem] = []
        enums = self.construct.encmapping
        for name, value in enums.items():
            items.append(EnumItem(name=name, value=value))
        return items

    def get_enum_item_from_obj(self) -> EnumItem:
        """ Get items to select in the ComboBox """
        obj = self.obj
        if isinstance(obj, int):
            if obj in self.construct.decmapping:
                return EnumItem(
                    name=str(self.construct.decmapping[obj]), value=int(obj)
                )
            else:
                return EnumItem(name=str(obj), value=int(obj))
        else:
            return EnumItem(name=str(obj), value=int(self.construct.encmapping[obj]))

    def conv_str_to_obj(self, s: str) -> Any:
        """ Convert string (enum name or integer value) to object """
        try:
            if s in self.construct.encmapping:
                value = self.construct.encmapping[s]  # type: ignore
            else:
                value = str_to_int(s)

            if value in self.construct.decmapping:
                new_obj = self.construct.decmapping[value]
            else:
                new_obj = cs.EnumInteger(value)
        except Exception:
            new_obj = s  # this will probably result in a binary-build-error
        return new_obj


# EntryFlagsEnum ######################################################################################################
class EntryFlagsEnum(EntrySubconstruct):
    construct: "cs.FlagsEnum"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cs.FlagsEnum",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        return super().typ_str + " as Flags"

    @property
    def obj_str(self) -> str:
        return self.conv_obj_to_str(self.obj)

    def create_obj_panel(self, parent) -> ObjPanel:
        return ObjPanel_FlagsEnum(parent, self)

    def conv_obj_to_str(self, obj: Any) -> str:
        try:
            val = 0
            flags = []
            for flag in self.construct.flags.keys():
                if obj[flag] is True:
                    flags.append(flag)
                    val |= self.construct.flags[flag]
            return f"{int_to_str(self.model.integer_format, int(val))} ({' | '.join(flags)})"
        except Exception:
            return str(self.obj)

    def get_flagsenum_items_from_obj(self) -> t.List[FlagsEnumItem]:
        """ Get items to show in the ComboBox """
        items: t.List[FlagsEnumItem] = []
        flags = self.construct.flags
        obj = self.obj
        for flag in flags.keys():
            items.append(
                FlagsEnumItem(name=str(flag), value=flags[flag], checked=obj[flag])
            )
        return items

    def conv_flagsenum_items_to_obj(self, items: t.List[FlagsEnumItem]) -> Any:
        """ Convert flagsenum items to object """
        new_obj = cs.Container()
        for item in items:
            if item.checked:
                new_obj[item.name] = True
            else:
                new_obj[item.name] = False
        return new_obj


# EntryTEnum ##########################################################################################################
class EntryTEnum(EntrySubconstruct):
    construct: "cst.TEnum[Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cst.TEnum[Any]",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        return super().typ_str + " as Enum"

    @property
    def obj_str(self) -> str:
        try:
            return f"{int_to_str(self.model.integer_format, int(self.obj.value))} ({str(self.obj)})"
        except Exception:
            return str(self.obj)

    def create_obj_panel(self, parent) -> ObjPanel:
        return ObjPanel_Enum(parent, self)

    def get_enum_items(self) -> t.List[EnumItem]:
        """ Get items to show in the ComboBox """
        items: t.List[EnumItem] = []
        enum_type: t.Type[cst.EnumBase] = self.construct.enum_type
        for e in enum_type:
            items.append(EnumItem(name=str(e), value=e.value))
        return items

    def get_enum_item_from_obj(self) -> EnumItem:
        """ Get items to select in the ComboBox """
        obj: cst.EnumBase = self.obj
        return EnumItem(name=str(obj), value=obj.value)

    def conv_str_to_obj(self, s: str) -> Any:
        """ Convert string (enum name or integer value) to object """
        enum_type: t.Type[cst.EnumBase] = self.construct.enum_type
        try:
            try:
                new_obj = enum_type[s]
            except KeyError:
                value = str_to_int(s)
                new_obj = enum_type(value)
        except Exception:
            new_obj = s  # this will probably result in a binary-build-error
        return new_obj


# EntryTFlagsEnum #####################################################################################################
class EntryTFlagsEnum(EntrySubconstruct):
    construct: "cst.TFlagsEnum[Any]"

    def __init__(
        self,
        model: "construct_editor.ConstructEditorModel",
        parent: Optional["EntryConstruct"],
        construct: "cst.TFlagsEnum[Any]",
    ):
        super().__init__(model, parent, construct)

    @property
    def typ_str(self) -> str:
        return super().typ_str + " as Flags"

    @property
    def obj_str(self) -> str:
        return self.conv_obj_to_str(self.obj)

    def create_obj_panel(self, parent) -> ObjPanel:
        return ObjPanel_FlagsEnum(parent, self)

    def conv_obj_to_str(self, obj: Any) -> str:
        try:
            flags = []
            for flag in self.construct.enum_type:
                if flag & obj == flag:
                    flags.append(flag.name)

            return f"{int_to_str(self.model.integer_format, int(obj))} ({' | '.join(flags)})"
        except Exception:
            return str(self.obj)

    def get_flagsenum_items_from_obj(self) -> t.List[FlagsEnumItem]:
        """ Get items to show in the ComboBox """
        items: t.List[FlagsEnumItem] = []
        enum_type: t.Type[cst.FlagsEnumBase] = self.construct.enum_type
        obj: cst.FlagsEnumBase = self.obj
        for flag in enum_type:
            items.append(
                FlagsEnumItem(
                    name=str(flag),
                    value=flag.value,
                    checked=True if flag & obj == flag else False,
                )
            )
        return items

    def conv_flagsenum_items_to_obj(self, items: t.List[FlagsEnumItem]) -> Any:
        """ Convert flagsenum items to object """
        enum_type: t.Type[cst.FlagsEnumBase] = self.construct.enum_type
        new_obj = enum_type(0)
        for item in items:
            if item.checked:
                new_obj |= enum_type(item.value)
        return new_obj


# #####################################################################################################################
# Entry Mapping #######################################################################################################
# #####################################################################################################################
@dataclasses.dataclass
class ClassEntryMapping:
    constr_class: Type["cs.Construct[Any, Any]"]
    entry: Type[EntryConstruct]


@dataclasses.dataclass
class SigletonEntryMapping:
    constr_instance: "cs.Construct[Any, Any]"
    entry: Type[EntryConstruct]


construct_entry_mapping: t.List[t.Union[ClassEntryMapping, SigletonEntryMapping]] = [
    # #########################################################################
    # wrapper from: construct #################################################
    # #########################################################################
    # bytes and bits ############################
    ClassEntryMapping(cs.Bytes, EntryBytes),
    SigletonEntryMapping(cs.GreedyBytes, EntryBytes),
    # cs.Bitwise
    # cs.Bytewise
    #
    # integers and floats #######################
    ClassEntryMapping(cs.FormatField, EntryFormatField),
    ClassEntryMapping(cs.BytesInteger, EntryBytesInteger),
    ClassEntryMapping(cs.BitsInteger, EntryBitsInteger),
    #
    # strings ###################################
    # cs.StringEncoded
    #
    # mappings ##################################
    # cs.Flag
    ClassEntryMapping(cs.Enum, EntryEnum),
    ClassEntryMapping(cs.FlagsEnum, EntryFlagsEnum),
    # cs.Mapping
    #
    # structures and sequences ##################
    ClassEntryMapping(cs.Struct, EntryStruct),
    # cs.Sequence
    #
    # arrays ranges and repeaters ###############
    ClassEntryMapping(cs.Array, EntryArray),
    ClassEntryMapping(cs.GreedyRange, EntryArray),
    # cs.RepeatUntil
    #
    # specials ##################################
    ClassEntryMapping(cs.Renamed, EntryRenamed),
    #
    # miscellaneous #############################
    ClassEntryMapping(cs.Const, EntryTransparentSubcon),
    ClassEntryMapping(cs.Computed, EntryComputed),
    # cs.Index
    # cs.Rebuild
    ClassEntryMapping(cs.Default, EntryTransparentSubcon),
    # cs.Check
    # cs.Error
    # cs.FocusedSeq
    # cs.Pickled
    # cs.Numpy
    # cs.NamedTuple
    ClassEntryMapping(cs.TimestampAdapter, EntryTimestamp),
    # cs.Hex
    # cs.HexDump
    #
    # conditional ###############################
    # cs.Union
    # cs.Select
    ClassEntryMapping(cs.IfThenElse, EntryIfThenElse),
    ClassEntryMapping(cs.Switch, EntrySwitch),
    # cs.StopIf
    #
    # alignment and padding #####################
    ClassEntryMapping(cs.Padded, EntryTransparentSubcon),
    ClassEntryMapping(cs.Aligned, EntryTransparentSubcon),
    #
    # stream manipulation #######################
    ClassEntryMapping(cs.Pointer, EntryTransparentSubcon),
    ClassEntryMapping(cs.Peek, EntryPeek),
    ClassEntryMapping(cs.Seek, EntrySeek),
    SigletonEntryMapping(cs.Tell, EntryTell),
    SigletonEntryMapping(cs.Pass, EntryPass),
    # cs.Terminated
    #
    # tunneling and byte/bit swapping ###########
    ClassEntryMapping(cs.RawCopy, EntryRawCopy),
    # cs.Prefixed
    # cs.FixedSized
    # cs.NullTerminated
    # cs.NullStripped
    # cs.RestreamData
    ClassEntryMapping(cs.Transformed, EntryTransparentSubcon),
    ClassEntryMapping(cs.Restreamed, EntryTransparentSubcon),
    # cs.ProcessXor
    # cs.ProcessRotateLeft
    # cs.Checksum
    # cs.Compressed
    # cs.CompressedLZ4
    # cs.Rebuffered
    # #########################################################################
    #
    #
    # #########################################################################
    # wrapper from: construct_typing ##########################################
    # #########################################################################
    ClassEntryMapping(cst.DataclassStruct, EntryDataclassStruct),
    ClassEntryMapping(cst.TEnum, EntryTEnum),
    ClassEntryMapping(cst.TFlagsEnum, EntryTFlagsEnum),
    # #########################################################################
    #
    #
    # #########################################################################
    # wrapper from: construct_editor ##########################################
    # #########################################################################
    ClassEntryMapping(IncludeGuiMetaData, EntryTransparentSubcon),
    # #########################################################################
]


def create_entry_from_construct(
    model: "construct_editor.ConstructEditorModel",
    parent: Optional["EntryConstruct"],
    subcon: "cs.Construct[Any, Any]",
) -> "EntryConstruct":

    # at first check singleton mappings
    for mapping in construct_entry_mapping:
        if (
            isinstance(mapping, SigletonEntryMapping)
            and subcon is mapping.constr_instance
        ):
            return mapping.entry(model, parent, subcon)

    # then check class mappings
    for mapping in construct_entry_mapping:
        if isinstance(mapping, ClassEntryMapping) and isinstance(
            subcon, mapping.constr_class
        ):
            return mapping.entry(model, parent, subcon)

    # use fallback, if no mapping is found
    if isinstance(subcon, cs.Construct):
        return EntryConstruct(model, parent, subcon)

    raise ValueError("subcon type {} is not implemented".format(repr(subcon)))



import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


@dataclasses.dataclass
class ComputedTest(cst.DataclassMixin):
    type_int: int = cst.csfield(cs.Computed(lambda ctx: 50))
    type_float: float = cst.csfield(cs.Computed(lambda ctx: 80.0))
    type_bool: bool = cst.csfield(cs.Computed(lambda ctx: True))
    type_bytes: bytes = cst.csfield(cs.Computed(lambda ctx: bytes([0x00, 0xAB])))
    type_bytearray: bytearray = cst.csfield(
        cs.Computed(lambda ctx: bytearray([0x00, 0xAB, 0xFF]))
    )


constr = cst.DataclassStruct(ComputedTest)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={"Zeros": bytes([])},
)


import construct as cs
from . import GalleryItem


constr = cs.Struct(
    "value1" / cs.Int8sb,
    "pass" / cs.Pass,
    "value2" / cs.Int8sb,
)


gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Zeros": bytes(2),
        "1": b"12",
    },
)


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
import arrow
from . import GalleryItem


@dataclasses.dataclass
class Timestamps(cst.DataclassMixin):
    time_int8: arrow.Arrow = cst.csfield(
        cs.Timestamp(cs.Int8ul, unit=1, epoch=arrow.Arrow(2020, 1, 1))
    )
    time_int16: arrow.Arrow = cst.csfield(
        cs.Timestamp(cs.Int16ul, unit=1, epoch=arrow.Arrow(2019, 1, 1))
    )
    time_int32: arrow.Arrow = cst.csfield(
        cs.Timestamp(cs.Int32ul, unit=1, epoch=arrow.Arrow(2080, 1, 1))
    )


constr = cst.DataclassStruct(Timestamps)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Zeros": bytes(constr.sizeof()),
        "1": bytes([4, 4, 12, 4, 4, 4, 4]),
    },
)



# -*- coding: utf-8 -*-
import enum
from typing import Any, Callable, Dict, List, Optional

import construct as cs
import wx
import wx.dataview as dv
from construct_editor.helper import CallbackList
import textwrap

from construct_editor.helper.preprocessor import include_metadata
from construct_editor.helper.wrapper import (
    ObjPanel,
    EntryConstruct,
    create_entry_from_construct,
    get_gui_metadata,
    add_gui_metadata,
)


class RootObjChangedCallbackList(CallbackList[Callable[[Any], None]]):
    def fire(self, root_obj: Any):
        for listener in self:
            listener(root_obj)


class EntrySelectedCallbackList(CallbackList[Callable[[EntryConstruct], None]]):
    def fire(self, entry: EntryConstruct):
        for listener in self:
            listener(entry)


# #####################################################################################################################
# Object Renderer #####################################################################################################
# #####################################################################################################################
class ObjectRenderer(dv.DataViewCustomRenderer):
    def __init__(self):
        super().__init__(varianttype="PyObject", mode=dv.DATAVIEW_CELL_EDITABLE)
        self.entry: Optional[EntryConstruct] = None
        self.EnableEllipsize(wx.ELLIPSIZE_END)

    def SetValue(self, value):
        self.entry = value
        return True

    def GetValue(self):
        return self.entry

    def GetSize(self):
        # Return the size needed to display the value.  The renderer
        # has a helper function we can use for measuring text that is
        # aware of any custom attributes that may have been set for
        # this item.
        obj_str = self.entry.obj_str if self.entry else ""
        size = self.GetTextExtent(obj_str)
        size += (2, 2)
        return size

    def Render(self, rect, dc, state):
        # And then finish up with this helper function that draws the
        # text for us, dealing with alignment, font and color
        # attributes, etc.
        obj_str = self.entry.obj_str if self.entry else ""
        self.RenderText(
            obj_str, 0, rect, dc, state  # x-offset  # wxDataViewCellRenderState flags
        )
        return True

    def ActivateCell(self, rect, model, item, col, mouseEvent):
        return False

    # The HasEditorCtrl, CreateEditorCtrl and GetValueFromEditorCtrl
    # methods need to be implemented if this renderer is going to
    # support in-place editing of the cell value, otherwise they can
    # be omitted.

    def HasEditorCtrl(self):
        return True

    def CreateEditorCtrl(self, parent, labelRect: wx.Rect, value: EntryConstruct):
        editor: ObjPanel = value.create_obj_panel(parent)
        editor.SetPosition(labelRect.Position)
        editor.SetSize(labelRect.Size)
        return editor

    def GetValueFromEditorCtrl(self, editor: ObjPanel):
        new_obj = editor.get_new_obj()
        # new_obj is passed in a dict, because subtypes of `int` (eg. `IntEnum`)
        # and `str` (eg. `cs.EnumIntegerString`) are converted to its base type
        # by the DVC...
        return {"new_obj": new_obj}


# #####################################################################################################################
# Context Menu ########################################################################################################
# #####################################################################################################################
class ContextMenu(wx.Menu):
    def __init__(
        self,
        parent: "ConstructEditor",
        model: "ConstructEditorModel",
        entry: Optional["EntryConstruct"],
    ):
        super(ContextMenu, self).__init__()
        self.parent = parent
        self.model = model

        item: wx.MenuItem = self.Append(wx.ID_COPY, "Copy\tCtrl+C")
        # self.Bind(wx.EVT_MENU, self.on_copy, id=item.Id)
        item.Enable(False)

        item: wx.MenuItem = self.Append(wx.ID_PASTE, "Paste\tCtrl+V")
        # self.Bind(wx.EVT_MENU, self.on_paste, id=item.Id)
        item.Enable(False)  # TODO:

        self.AppendSeparator()

        item: wx.MenuItem = self.Append(wx.ID_UNDO, "Undo\tCtrl+Z")
        self.Bind(
            wx.EVT_MENU,
            lambda event: self.model.command_processor.Undo(),
            id=item.Id,
        )
        item.Enable(self.model.command_processor.CanUndo())

        item: wx.MenuItem = self.Append(wx.ID_REDO, "Redo\tCtrl+Y")
        self.Bind(
            wx.EVT_MENU,
            lambda event: self.model.command_processor.Redo(),
            id=item.Id,
        )
        item.Enable(self.model.command_processor.CanRedo())

        self.AppendSeparator()

        item: wx.MenuItem = self.Append(wx.ID_ANY, "Expand All\tCtrl+E")
        self.Bind(wx.EVT_MENU, self.on_expand_all, id=item.Id)

        item: wx.MenuItem = self.Append(wx.ID_ANY, "Collapse All\tCtrl+W")
        self.Bind(wx.EVT_MENU, self.on_collapse_all, id=item.Id)

        self.AppendSeparator()

        item: wx.MenuItem = self.AppendCheckItem(wx.ID_ANY, "Hide Protected")
        self.Bind(wx.EVT_MENU, self.on_hide_protected, id=item.Id)
        item.Check(self.parent.hide_protected)
        self.hide_protected_mi = item

        self.AppendSeparator()

        item: wx.MenuItem = self.AppendRadioItem(wx.ID_ANY, "Dec")
        self.Bind(wx.EVT_MENU, self.on_intformat_dec, id=item.Id)
        self.intformat_dec_mi = item

        item: wx.MenuItem = self.AppendRadioItem(wx.ID_ANY, "Hex")
        self.Bind(wx.EVT_MENU, self.on_intformat_hex, id=item.Id)
        self.intformat_hex_mi = item

        if self.model.integer_format is IntegerFormat.Hex:
            self.intformat_hex_mi.Check(True)
        else:
            self.intformat_dec_mi.Check(True)

        # Add List with all currently shown list viewed items
        if len(model.list_viewed_entries) > 0:
            self.AppendSeparator()

            submenu = wx.Menu()
            self.submenu_map: Dict[Any, EntryConstruct] = {}
            for e in model.list_viewed_entries:
                name = "->".join(e.path)
                item: wx.MenuItem = submenu.AppendCheckItem(wx.ID_ANY, name)
                self.submenu_map[item.GetId()] = e
                self.Bind(wx.EVT_MENU, self.on_remove_list_viewed_item, item)
                item.Check(True)

            self.AppendSubMenu(submenu, "List Viewed Items")

        # Add additional items for this entry
        if entry is not None:
            entry.modify_context_menu(self)

    def on_expand_all(self, event):
        self.parent.expand_all()

    def on_collapse_all(self, event):
        self.parent.collapse_all()

    def on_hide_protected(self, event):
        checked = self.hide_protected_mi.IsChecked()
        self.parent.hide_protected = checked
        self.parent.reload()

    def on_intformat_dec(self, event: wx.CommandEvent):
        self.intformat_dec_mi.Check(True)
        self.model.integer_format = IntegerFormat.Dec
        self.parent.reload()
        event.Skip()

    def on_intformat_hex(self, event: wx.CommandEvent):
        self.intformat_hex_mi.Check(True)
        self.model.integer_format = IntegerFormat.Hex
        self.parent.reload()
        event.Skip()

    def on_remove_list_viewed_item(self, event: wx.CommandEvent):
        entry = self.submenu_map[event.GetId()]
        self.model.list_viewed_entries.remove(entry)
        self.parent.reload()


# #####################################################################################################################
# Construct Editor Model ##############################################################################################
# #####################################################################################################################


class IntegerFormat(enum.Enum):
    Dec = enum.auto()
    Hex = enum.auto()


class ConstructEditorColumn(enum.IntEnum):
    Name = 0
    Type = 1
    Value = 2


class ConstructEditorModel(dv.PyDataViewModel):
    """
    This model acts as a bridge between the DataViewCtrl and the dataclasses.
    This model provides these data columns:
        0. Name: string
        1. Type: string
        2. Value: string
    """

    def __init__(self, dvc: dv.DataViewCtrl):
        dv.PyDataViewModel.__init__(self)
        self.dvc = dvc
        self.root_entry: Optional["EntryConstruct"] = None
        self.root_obj: Optional[Any] = None

        self.hide_protected = True

        # List with all entries that have the list view enabled
        self.list_viewed_entries: List["EntryConstruct"] = []

        # The PyDataViewModel derives from both DataViewModel and from
        # DataViewItemObjectMapper, which has methods that help associate
        # data view items with Python objects. Normally a dictionary is used
        # so any Python object can be used as data nodes. If the data nodes
        # are weak-referencable then the objmapper can use a
        # WeakValueDictionary instead.
        self.UseWeakRefs(True)

        self.command_processor = wx.CommandProcessor()

        self.integer_format = IntegerFormat.Dec

    def get_column_count(self):
        """ Get the column count """
        return 3

    def get_column_name(self, selected_item, col):
        """ Get the name of the column. The column name depends on the selected item """
        entry = self.ItemToObject(selected_item)
        if not isinstance(entry, EntryConstruct):
            raise ValueError(f"{repr(entry)} is no valid entry")

        flat_subentry_list: List["EntryConstruct"] = []
        entry.create_flat_subentry_list(flat_subentry_list)

        return "->".join(flat_subentry_list[col].path)

    # #################################################################################################################
    # dv.PyDataViewModel Interface ####################################################################################
    # #################################################################################################################
    def GetColumnCount(self):
        # Report how many columns this model provides data for.
        return self.get_column_count()

    def GetChildren(self, parent, children):

        # The view calls this method to find the children of any node in the
        # control. There is an implicit hidden root node, and the top level
        # item(s) should be reported as children of this node. A List view
        # simply provides all items as children of this hidden root. A Tree
        # view adds additional items as children of the other items, as needed,
        # to provide the tree hierarchy.

        # If the parent item is invalid then it represents the hidden root
        # item, so we'll use the genre objects as its children and they will
        # end up being the collection of visible roots in our tree.

        if self.root_entry is None:
            return 0

        if not parent:
            entry = self.root_entry
            item = self.ObjectToItem(entry)
            entry.dvc_item = item
            children.append(item)
            return 1

        parent_entry = self.ItemToObject(parent)
        if not isinstance(parent_entry, EntryConstruct):
            raise ValueError(f"{repr(parent_entry)} is no valid entry")

        for entry in parent_entry.subentries:
            name = entry.name
            if (self.hide_protected == True) and (name.startswith("_") or name == ""):
                continue
            item = self.ObjectToItem(entry)
            entry.dvc_item = item
            children.append(item)
        return len(children)

    def IsContainer(self, item):
        # Return True if the item has children, False otherwise.

        # The hidden root is a container
        if not item:
            return True

        # check if the entry is a container
        entry: EntryConstruct = self.ItemToObject(item)
        return entry.subentries is not None

    def HasContainerColumns(self, item):
        # True zurückgeben, damit in Containern auch in allen Spalten Werte angezeigt werden
        return True

    def GetParent(self, item):
        # Return the item which is this item's parent.

        # Root object
        if not item:
            return dv.NullDataViewItem

        # return parent of entry
        entry: EntryConstruct = self.ItemToObject(item)
        if entry.parent is None:
            return dv.NullDataViewItem
        else:
            return entry.parent.dvc_item

    def GetValue(self, item: dv.DataViewItem, col: int):
        # Return the value to be displayed for this item and column.

        entry = self.ItemToObject(item)
        if not isinstance(entry, EntryConstruct):
            raise ValueError(f"{repr(entry)} is no valid entry")

        if col == ConstructEditorColumn.Name:
            return entry.name
        if col == ConstructEditorColumn.Type:
            return entry.typ_str
        if col == ConstructEditorColumn.Value:
            return entry

        if (entry.parent is None) or (entry.parent not in self.list_viewed_entries):
            return ""

        # flatten the hierarchical structure to a list
        col = col - len(ConstructEditorColumn)

        flat_subentry_list: List["EntryConstruct"] = []
        entry.create_flat_subentry_list(flat_subentry_list)
        if len(flat_subentry_list) > col:
            return flat_subentry_list[col].obj_str
        else:
            return ""

    def SetValue(self, value: Any, item: dv.DataViewItem, col: int):
        if col != ConstructEditorColumn.Value:
            raise ValueError(f"col={col} cannot be modified")

        entry: EntryConstruct = self.ItemToObject(item)
        if not isinstance(entry, EntryConstruct):
            raise ValueError(f"{repr(entry)} is no valid entry")

        # new_obj is passed in a dict, because subtypes of `int` (eg. `IntEnum`)
        # and `str` (eg. `cs.EnumIntegerString`) are converted to its base type
        # by the DVC...
        new_obj = value["new_obj"]

        # get the current object
        obj = entry.obj

        # link the metadata from the current object to the new one
        metadata = get_gui_metadata(obj)
        if metadata is not None:
            new_obj = add_gui_metadata(new_obj, metadata)

        model = self

        class Cmd(wx.Command):
            def __init__(self):
                super().__init__(True, f"Value '{entry.path[-1]}' changed")

            def Do(self):
                self._obj_backup = obj
                entry.obj = new_obj
                model.ItemChanged(entry.dvc_item)
                return True

            def Undo(self):
                entry.obj = self._obj_backup
                model.ItemChanged(entry.dvc_item)
                return True

        self.command_processor.Submit(Cmd())
        return True

    def GetAttr(self, item, col, attr):
        entry = self.ItemToObject(item)

        if entry is self.root_entry:
            attr.SetColour("blue")
            attr.SetBold(True)
            return True

        return False


# #####################################################################################################################
# Construct Editor ####################################################################################################
# #####################################################################################################################
class ConstructEditor(wx.Panel):
    def __init__(
        self,
        parent,
        construct: cs.Construct,
    ):
        super().__init__(parent)

        vsizer = wx.BoxSizer(wx.VERTICAL)

        # Create DataViewCtrl
        self._dvc = dv.DataViewCtrl(
            self,
            wx.ID_ANY,
            wx.DefaultPosition,
            wx.DefaultSize,
            style=dv.DV_VERT_RULES | dv.DV_ROW_LINES,
            name="construct_editor",
        )
        vsizer.Add(self._dvc, 3, wx.ALL | wx.EXPAND, 5)

        # Create Model of DataViewCtrl
        self._model = ConstructEditorModel(self._dvc)
        self._dvc.AssociateModel(self._model)

        # Create InfoBars
        self._parse_error_info_bar = wx.InfoBar(self)
        vsizer.Add(self._parse_error_info_bar, 0, wx.EXPAND)

        self._build_error_info_bar = wx.InfoBar(self)
        vsizer.Add(self._build_error_info_bar, 0, wx.EXPAND)

        # create status bar
        self._status_bar = wx.StatusBar(
            self,
            style=wx.STB_SHOW_TIPS | wx.STB_ELLIPSIZE_END | wx.FULL_REPAINT_ON_RESIZE,
        )
        self._status_bar.SetFieldsCount(2)
        self._status_bar.SetStatusStyles(
            [wx.SB_NORMAL, wx.SB_FLAT]
        )  # remove vertical line after the last field
        self._status_bar.SetStatusWidths([-2, -1])
        vsizer.Add(self._status_bar, 0, wx.ALL | wx.EXPAND, 0)

        self.SetSizer(vsizer)
        self.Layout()

        # Connect Events
        self._dvc.Bind(
            dv.EVT_DATAVIEW_SELECTION_CHANGED,
            self._on_dvc_selection_changed,
            id=wx.ID_ANY,
        )
        self._dvc.Bind(dv.EVT_DATAVIEW_ITEM_VALUE_CHANGED, self._on_dvc_value_changed)
        self._dvc.Bind(dv.EVT_DATAVIEW_ITEM_CONTEXT_MENU, self._on_dvc_right_clicked)
        self._dvc.Bind(dv.EVT_DATAVIEW_ITEM_EXPANDED, self._on_dvc_item_expanded)
        self._dvc.Bind(dv.EVT_DATAVIEW_ITEM_COLLAPSED, self._on_dvc_item_collapsed)

        self._dvc_main_window: wx.Window = self._dvc.GetMainWindow()
        self._dvc_main_window.Bind(wx.EVT_MOTION, self._on_dvc_motion)
        self._dvc_main_window.Bind(wx.EVT_KEY_DOWN, self._on_dvc_key_down)
        self._last_motion_obj: Optional[EntryConstruct] = None

        self._dvc_header = self._dvc.FindWindowByName("wxMSWHeaderCtrl")
        if self._dvc_header is not None:
            self._dvc_header.Bind(wx.EVT_MOTION, self._on_dvc_header_motion)

        self.on_entry_selected = EntrySelectedCallbackList()
        self.on_root_obj_changed = RootObjChangedCallbackList()
        self.construct = construct

    def reload(self):
        """ Reload the ConstructEditor, while remaining expaned elements and selection """
        try:
            self.Freeze()

            # reload dvc columns
            self._reload_dvc_columns()

            # save selection
            selections = self._dvc.GetSelections()

            # clear the dvc.
            # unfortunately the selection and expanded items get lost... so we have to save and restore it manually
            self._model.Cleared()
            self._clear_status_bar()

            # expand everything
            self._model.root_entry.dvc_item_restore_expansion()

            # restore selection
            self._dvc.SetSelections(selections)

        finally:
            self.Thaw()

    def parse(self, binary: bytes, **contextkw: Any):
        """ Parse binary data to struct. """
        try:
            self._model.root_obj = self._construct.parse(binary, **contextkw)
            self._parse_error_info_bar.Dismiss()
        except Exception as e:
            self._parse_error_info_bar.ShowMessage(
                f"Error while parsing binary data: {type(e).__name__}\n{str(e)}",
                wx.ICON_WARNING,
            )
            self._model.root_obj = None
        # clear all commands, when new data is set from external
        self._model.command_processor.ClearCommands()
        self.reload()

    def build(self, **contextkw: Any) -> bytes:
        """ Build binary data from struct. """
        try:
            binary = self._construct.build(self.root_obj, **contextkw)
            self._build_error_info_bar.Dismiss()
        except Exception as e:
            self._build_error_info_bar.ShowMessage(
                f"Error while building binary data: {type(e).__name__}\n{str(e)}",
                wx.ICON_WARNING,
            )
            raise e

        # parse the build binary, so that constructs that parses from nothing are shown correctly (eg. cs.Peek)
        wx.CallAfter(lambda: self.parse(binary, **contextkw))

        return binary

    # Property: construct #####################################################
    @property
    def construct(self) -> cs.Construct:
        """ Construct that is used for displaying. """
        return self._construct

    @construct.setter
    def construct(self, val: cs.Construct):
        # add root name, is none is available
        if val.name is None:
            val = "root" / val

        # modify the copied construct, so that each item also includes metadata for the GUI
        self._construct = include_metadata(val)

        # create entry from the construct
        self._model.root_entry = create_entry_from_construct(
            self._model, None, self._construct
        )

        self._model.list_viewed_entries.clear()

    # Property: root_obj ######################################################
    @property
    def root_obj(self) -> Any:
        return self._model.root_obj

    # Property: hide_protected ################################################
    @property
    def hide_protected(self) -> bool:
        """
        Hide protected members.
        A protected member starts with an undescore (_)
        """
        return self._model.hide_protected

    @hide_protected.setter
    def hide_protected(self, value: bool):
        self._model.hide_protected = value
        self.reload()

    # expand_all ##############################################################
    def expand_all(self):
        """
        Expand all Entries
        """

        def dvc_expand(entry: EntryConstruct):
            if entry.subentries is not None:
                if entry.dvc_item is not None:
                    self._dvc.Expand(entry.dvc_item)
                for sub_entry in entry.subentries:
                    dvc_expand(sub_entry)

        if self._model.root_entry:
            dvc_expand(self._model.root_entry)

    # expand_level ############################################################
    def expand_level(self, level: int):
        """
        Expand all Entries to Level ... (0=root level)
        """

        def dvc_expand(entry: EntryConstruct, current_level: int):
            subentries = entry.subentries
            dvc_item = entry.dvc_item
            if subentries is not None:
                if dvc_item is not None:
                    self._dvc.Expand(dvc_item)
                if current_level < level:
                    for sub_entry in subentries:
                        dvc_expand(sub_entry, current_level + 1)

        if self._model.root_entry:
            dvc_expand(self._model.root_entry, 1)

    # collapse_all ############################################################
    def collapse_all(self):
        """
        Collapse all Entries
        """

        def dvc_collapse(entry: EntryConstruct):
            subentries = entry.subentries
            dvc_item = entry.dvc_item
            if subentries is not None:
                for sub_entry in subentries:
                    dvc_collapse(sub_entry)
                if dvc_item is not None:
                    self._dvc.Collapse(dvc_item)

        if self._model.root_entry:
            dvc_collapse(self._model.root_entry)

        # expand the root entry again
        self.expand_level(1)

    # Internals ###############################################################
    def _reload_dvc_columns(self):
        """ Reload the dvc columns """
        self._dvc.ClearColumns()

        self._dvc.AppendTextColumn("Name", ConstructEditorColumn.Name, width=160)
        self._dvc.AppendTextColumn("Type", ConstructEditorColumn.Type, width=90)
        # self._dvc.AppendTextColumn("Value", ConstructEditorColumn.Value, width=200)

        renderer = ObjectRenderer()
        col = dv.DataViewColumn(
            "Value", renderer, ConstructEditorColumn.Value, width=200
        )
        col.Alignment = wx.ALIGN_LEFT
        self._dvc.AppendColumn(col)

        list_cols = 0
        for list_viewed_entry in self._model.list_viewed_entries:
            for subentry in list_viewed_entry.subentries:
                flat_list = []
                subentry.create_flat_subentry_list(flat_list)
                list_cols = max(list_cols, len(flat_list))

        for list_col in range(list_cols):
            self._dvc.AppendTextColumn(
                str(list_col), len(ConstructEditorColumn) + list_col
            )

    def _rename_dvc_columns(self, entry: EntryConstruct):
        """ Rename the dvc columns """

        flat_list: List["EntryConstruct"] = []
        if (entry.parent is not None) and (
            entry.parent in self._model.list_viewed_entries
        ):
            entry.create_flat_subentry_list(flat_list)

        list_cols = self._dvc.GetColumnCount() - len(ConstructEditorColumn)
        for list_col in range(list_cols):
            dvc_column: dv.DataViewColumn = self._dvc.GetColumn(
                len(ConstructEditorColumn) + list_col
            )
            if list_col < len(flat_list):
                path = flat_list[list_col].path
                path = path[len(entry.path) :]  # remove the path from the parent
                dvc_column.SetTitle("->".join(path))
            else:
                dvc_column.SetTitle(str(list_col))

    def _on_dvc_selection_changed(self, event):
        """
        This method is called, if the selection in the dvc has changed.

        Then the infos of the new selected entry is shown.
        """
        item = self._dvc.GetSelection()
        if item.ID is not None:
            entry: EntryConstruct = self._model.ItemToObject(item)
            self._refresh_status_bar(entry)

            self.on_entry_selected.fire(entry)

            self._rename_dvc_columns(entry)

        else:
            self._clear_status_bar()

    def _refresh_status_bar(self, entry: EntryConstruct):
        self._status_bar.SetStatusText("->".join(entry.path), 0)
        bytes_info = ""
        metadata = entry.obj_metadata
        if metadata is not None:
            start = metadata.offset_start
            end = metadata.offset_end - 1
            size = end - start + 1
            if size > 0:
                bytes_info = f"Bytes: {start}-{end} ({size})"
        self._status_bar.SetStatusText(bytes_info, 1)

    def _clear_status_bar(self):
        self._status_bar.SetStatusText("", 0)
        self._status_bar.SetStatusText("", 1)

    def _on_dvc_value_changed(self, event: dv.DataViewEvent):
        """ This method is called, if a value in the dvc has changed. """
        self.on_root_obj_changed.fire(self._model.root_obj)

    def _on_dvc_motion(self, event: wx.MouseEvent):
        # this is a mouse event, so we have to calculate the position of
        # the item where the mouse is manually.
        pos = event.GetPosition()
        pos += (0, self._dvc_header.Size.Height)  # correct the dvc header
        item, col = self._dvc.HitTest(pos)
        if item.GetID() is None:
            self._dvc_main_window.SetToolTip("")
            return
        obj: EntryConstruct = self._model.ItemToObject(item)

        if col.ModelColumn == ConstructEditorColumn.Name:
            # only set tooltip if the obj changed. this prevents flickering
            if self._last_motion_obj is not obj:
                self._dvc_main_window.SetToolTip(textwrap.dedent(obj.docs).strip())
            self._last_motion_obj = obj
        else:
            self._dvc_main_window.SetToolTip("")
            self._last_motion_obj = None

    def _on_dvc_header_motion(self, event: wx.MouseEvent):
        event.Skip()  # TODO: Create Tooltip for DVC-Header

    def _on_dvc_right_clicked(self, event: dv.DataViewEvent):
        """
        This method is called, the dvc ist right clicked

        Then a context menu is created
        """
        item = event.GetItem()
        entry: Optional["EntryConstruct"]
        if item.ID is not None:
            entry = self._model.ItemToObject(item)
        else:
            entry = None
        self.PopupMenu(ContextMenu(self, self._model, entry), event.GetPosition())

    def _on_dvc_key_down(self, event: wx.KeyEvent):
        # Ctrl+E
        if event.ControlDown() and event.GetKeyCode() == ord("E"):
            self.expand_all()

        # Ctrl+W
        if event.ControlDown() and event.GetKeyCode() == ord("W"):
            self.collapse_all()

        # Ctrl+Z
        if event.ControlDown() and event.GetKeyCode() == ord("Z"):
            self._model.command_processor.Undo()

        # Ctrl+Y
        elif event.ControlDown() and event.GetKeyCode() == ord("Y"):
            self._model.command_processor.Redo()

        # Ctrl+C
        elif event.ControlDown() and event.GetKeyCode() == ord("C"):
            pass  # TODO

        # Ctrl+V
        elif event.ControlDown() and event.GetKeyCode() == ord("V"):
            pass  # TODO

        else:
            event.Skip()

    def _on_dvc_item_expanded(self, event: dv.DataViewEvent):
        item = event.GetItem()
        if item.ID is None:
            return
        entry: EntryConstruct = self._model.ItemToObject(item)
        entry.dvc_item_expanded = True

    def _on_dvc_item_collapsed(self, event: dv.DataViewEvent):
        item = event.GetItem()
        if item.ID is None:
            return
        entry: EntryConstruct = self._model.ItemToObject(item)
        entry.dvc_item_expanded = False


import dataclasses
import construct as cs
import construct_typed as cst
import typing as t
import copy


@dataclasses.dataclass
class GuiMetaData:
    offset_start: int
    offset_end: int
    length: int
    construct: cs.Construct
    index: t.Optional[int]
    context: "cs.Context"


class IntWithGuiMetadata(int):
    pass


class FloatWithGuiMetadata(float):
    pass


class BytesWithGuiMetadata(bytes):
    pass


class BytearrayWithGuiMetadata(bytearray):
    pass


class StrWithGuiMetadata(str):
    pass


class NoneWithGuiMetadata:
    pass


def get_gui_metadata(obj: t.Any) -> t.Optional[GuiMetaData]:
    """ Get the GUI metadata if they are available """
    try:
        return obj.__gui_metadata  # type: ignore
    except Exception:
        return None


def add_gui_metadata(obj: t.Any, gui_metadata: GuiMetaData) -> t.Any:
    """ Append the private field "__gui_metadata" to an object """
    try:
        obj.__gui_metadata = gui_metadata  # type: ignore
    except AttributeError:
        if isinstance(obj, int):
            obj = IntWithGuiMetadata(obj)
            obj.__gui_metadata = gui_metadata  # type: ignore
        elif isinstance(obj, float):
            obj = FloatWithGuiMetadata(obj)
            obj.__gui_metadata = gui_metadata  # type: ignore
        elif isinstance(obj, bytes):
            obj = BytesWithGuiMetadata(obj)
            obj.__gui_metadata = gui_metadata  # type: ignore
        elif isinstance(obj, bytearray):
            obj = BytearrayWithGuiMetadata(obj)
            obj.__gui_metadata = gui_metadata  # type: ignore
        elif isinstance(obj, str):
            obj = StrWithGuiMetadata(obj)
            obj.__gui_metadata = gui_metadata  # type: ignore
        elif obj is None:
            obj = NoneWithGuiMetadata()
            obj.__gui_metadata = gui_metadata  # type: ignore
        else:
            raise ValueError(f"add_gui_metadata dont work with type of {type(obj)}")
    return obj


class IncludeGuiMetaData(cs.Subconstruct):
    """ Include GUI metadata to the parsed object """

    def _parse(self, stream, context, path):
        # Get Index if we are in an Array, Sequence, ...
        if hasattr(context, "_index"):
            index = context._index
        else:
            index = None

        offset_start = cs.stream_tell(stream, path)
        obj = self.subcon._parsereport(stream, context, path)  # type: ignore
        offset_end = cs.stream_tell(stream, path)

        gui_metadata = GuiMetaData(
            offset_start=offset_start,
            offset_end=offset_end,
            length=offset_end - offset_start,
            construct=self.subcon,
            index=index,
            context=context,
        )

        return add_gui_metadata(obj, gui_metadata)

    def _build(self, obj, stream, context, path):
        buildret = self.subcon._build(obj, stream, context, path)  # type: ignore
        return obj


def include_metadata(
    constr: "cs.Construct[t.Any, t.Any]",
) -> "cs.Construct[t.Any, t.Any]":
    """
    Surrond all named entries of a construct with offsets, so that
    we know the offset in the byte-stream and the length
    """
    if isinstance(
        constr,
        (
            cs.BytesInteger,
            cs.BitsInteger,
            cs.Bytes,
            cs.FormatField,
            cs.BytesInteger,
            cs.BitsInteger,
            cs.Computed,
            cs.Check,
            cs.StopIf,
            cst.TEnum,
            cs.Enum,
            cs.FlagsEnum,
            cs.TimestampAdapter,
            cs.Seek,
        ),
    ):
        return IncludeGuiMetaData(constr)

    elif isinstance(constr, (cs.Restreamed, cs.Transformed)):
        # these constructs manipulate the stream, so the offsets of all nested subcons are wrong,
        # so we dont apply metadata recursivly
        return IncludeGuiMetaData(constr)

    elif isinstance(constr, (cs.Struct)):
        constr = copy.copy(constr)  # constr is modified, so we have to make a copy
        new_subcons = []
        for subcon in constr.subcons:
            new_subcons.append(include_metadata(subcon))
        constr.subcons = new_subcons
        return IncludeGuiMetaData(constr)

    elif isinstance(constr, (cs.Array, cs.GreedyRange)):
        constr = copy.copy(constr)  # constr is modified, so we have to make a copy
        constr.subcon = include_metadata(constr.subcon)
        return IncludeGuiMetaData(constr)

    elif isinstance(constr, cs.IfThenElse):
        constr = copy.copy(constr)  # constr is modified, so we have to make a copy
        constr.thensubcon = include_metadata(constr.thensubcon)
        constr.elsesubcon = include_metadata(constr.elsesubcon)
        return IncludeGuiMetaData(constr)

    elif isinstance(constr, cs.Switch):
        constr = copy.copy(constr)  # constr is modified, so we have to make a copy
        new_cases = {}
        for key, subcon in constr.cases.items():
            new_cases[key] = include_metadata(subcon)
        constr.cases = new_cases
        if constr.default is not None:
            constr.default = include_metadata(constr.default)
        return IncludeGuiMetaData(constr)

    elif isinstance(
        constr,
        (
            cs.Tunnel,
            cs.Renamed,
            cs.Const,
            cs.Rebuild,
            cs.Default,
            cs.Padded,
            cs.Aligned,
            cs.Prefixed,
            cs.Pointer,
            cs.Peek,
            cs.NullStripped,
            cs.FixedSized,
        ),
    ):
        constr = copy.copy(constr)  # constr is modified, so we have to make a copy
        constr.subcon = include_metadata(constr.subcon)  # type: ignore
        return constr

    elif isinstance(constr, cst.DataclassStruct):
        constr = copy.copy(constr)  # constr is modified, so we have to make a copy
        constr.subcon = include_metadata(constr.subcon)  # type: ignore
        return IncludeGuiMetaData(constr)

    elif isinstance(
        constr,
        (
            cs.ExprAdapter,
            cs.Adapter,
            type(cs.GreedyBytes),
            type(cs.VarInt),
            type(cs.ZigZag),
            type(cs.Flag),
            type(cs.Index),
            type(cs.Error),
            type(cs.Pickled),
            type(cs.Numpy),
            type(cs.Tell),
            type(cs.Pass),
            type(cs.Terminated),
        ),  # type: ignore
    ):
        return IncludeGuiMetaData(constr)

    # TODO:
    # # Grouping:
    # - Sequence
    # - FocusedSeq
    # - Union
    # - Select
    # - LazyStruct

    # # Grouping lists:
    # - Array
    # - GreedyRange
    # - RepeatUntil

    # # Special:
    # - Pointer
    # - RawCopy
    # - Restreamed
    # - Transformed
    # - RestreamData
    raise ValueError(f"construct of type '{constr}' is not supported")



import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


@dataclasses.dataclass
class ArrayTest(cst.DataclassMixin):
    simple_static: t.List[int] = cst.csfield(cs.Array(5, cs.Int8ub))

    simple_dynamic_len: int = cst.csfield(
        cs.Int8ub, doc="Length of the simple dynamic array"
    )
    simple_dynamic: t.List[int] = cst.csfield(
        cs.Array(cs.this.simple_dynamic_len, cs.Int8ub)
    )

    @dataclasses.dataclass
    class Entry(cst.DataclassMixin):
        id: int = cst.csfield(cs.Int8ub)
        width: int = cst.csfield(cs.Int8ub)
        height: int = cst.csfield(cs.Int8ub)

    struct_static: t.List[int] = cst.csfield(cs.Array(3, cst.DataclassStruct(Entry)))

    struct_dynamic_len: int = cst.csfield(
        cs.Int8ub, doc="Length of the struct dynamic array"
    )
    struct_dynamic: t.List[int] = cst.csfield(
        cs.Array(cs.this.struct_dynamic_len, cst.DataclassStruct(Entry))
    )


constr = cst.DataclassStruct(ArrayTest)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "Zeros": bytes(5 + 1 + 3 * 3 + 1),
        "1": (
            bytes([1, 2, 3, 4, 5, 4, 1, 2, 3, 4])
            + bytes([0xA0, 0xA1, 0xA2, 0xB0, 0xB1, 0xB2, 0xC0, 0xC1, 0xC2])
            + bytes([2, 0xA0, 0xA1, 0xA2, 0xB0, 0xB1, 0xB2])
        ),
        "2": (
            bytes([1, 2, 3, 4, 5, 8, 7, 6, 5, 4, 3, 2, 1, 0])
            + bytes([0xA0, 0xA1, 0xA2, 0xB0, 0xB1, 0xB2, 0xC0, 0xC1, 0xC2])
            + bytes([3, 0xA0, 0xA1, 0xA2, 0xB0, 0xB1, 0xB2, 0xC0, 0xC1, 0xC2])
        ),
    },
)


import construct as cs
import construct_typed as cst
import dataclasses
import typing as t
from . import GalleryItem


class Permission(cst.FlagsEnumBase):
    R = 4
    W = 2
    X = 1
    RWX = 7


class Day(cst.FlagsEnumBase):
    Monday = 1 << 0
    Tuesday = 1 << 1
    Wednesday = 1 << 2
    Thursday = 1 << 3
    Friday = 1 << 4
    Saturday = 1 << 5
    Sunday = 1 << 6


class LongList(cst.FlagsEnumBase):
    Entry0 = 1 << 0
    Entry1 = 1 << 1
    Entry2 = 1 << 2
    Entry3 = 1 << 3
    Entry4 = 1 << 4
    Entry5 = 1 << 5
    Entry6 = 1 << 6
    Entry7 = 1 << 7
    Entry8 = 1 << 8
    Entry9 = 1 << 9
    Entry10 = 1 << 10
    Entry11 = 1 << 11
    Entry12 = 1 << 12
    Entry13 = 1 << 13
    Entry14 = 1 << 14
    Entry15 = 1 << 15
    Entry16 = 1 << 16
    Entry17 = 1 << 17
    Entry18 = 1 << 18
    Entry19 = 1 << 19
    Entry20 = 1 << 20
    Entry21 = 1 << 21
    Entry22 = 1 << 22
    Entry23 = 1 << 23
    Entry24 = 1 << 24
    Entry25 = 1 << 25
    Entry26 = 1 << 26
    Entry27 = 1 << 27
    Entry28 = 1 << 28
    Entry29 = 1 << 29
    Entry30 = 1 << 30
    Entry31 = 1 << 31
    Entry32 = 1 << 32
    Entry33 = 1 << 33
    Entry34 = 1 << 34
    Entry35 = 1 << 35
    Entry36 = 1 << 36
    Entry37 = 1 << 37
    Entry38 = 1 << 38
    Entry39 = 1 << 39
    Entry40 = 1 << 40
    Entry41 = 1 << 41
    Entry42 = 1 << 42
    Entry43 = 1 << 43
    Entry44 = 1 << 44
    Entry45 = 1 << 45
    Entry46 = 1 << 46
    Entry47 = 1 << 47
    Entry48 = 1 << 48
    Entry49 = 1 << 49
    Entry50 = 1 << 50


@dataclasses.dataclass
class FlagsEnumTest(cst.DataclassMixin):
    permissions: Permission = cst.csfield(cst.TFlagsEnum(cs.Int8ul, Permission))
    days: Day = cst.csfield(cst.TFlagsEnum(cs.Int8ul, Day))
    long_list: LongList = cst.csfield(cst.TFlagsEnum(cs.Int64ul, LongList))


constr = cst.DataclassStruct(FlagsEnumTest)

gallery_item = GalleryItem(
    construct=constr,
    example_binarys={
        "2": constr.build(
            FlagsEnumTest(
                permissions=Permission.R | Permission.W,
                days=Day.Monday | Day.Sunday,
                long_list=LongList.Entry49 | LongList.Entry50,
            )
        ),
        "1": constr.build(
            FlagsEnumTest(
                permissions=Permission.R,
                days=Day.Monday,
                long_list=LongList.Entry0,
            )
        ),
        "Zeros": constr.build(
            FlagsEnumTest(
                permissions=Permission(0),
                days=Day(0),
                long_list=LongList(0),
            )
        ),
    },
)